import os
from unittest import mock

from api_fhir_r4.converters import PatientConverter

from api_fhir_r4.models import FHIRBaseObject
from api_fhir_r4.tests import PatientTestMixin


class PatientConverterTestCase(PatientTestMixin):

    __TEST_PATIENT_JSON_PATH = "/test/test_patient.json"

    def setUp(self):
        super(PatientConverterTestCase, self).setUp()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self._test_patient_json_representation = open(dir_path + self.__TEST_PATIENT_JSON_PATH).read()

    def test_to_fhir_obj(self):
        self.setUp()
        imis_insuree = self.create_test_imis_instance()
        fhir_patient = PatientConverter.to_fhir_obj(imis_insuree)
        self.verify_fhir_instance(fhir_patient)

    @mock.patch('insuree.models.Gender.objects')
    def test_to_imis_obj(self, mock_gender):
        self.setUp()
        mock_gender.get.return_value = self._TEST_GENDER

        fhir_patient = self.create_test_fhir_instance()
        imis_insuree = PatientConverter.to_imis_obj(fhir_patient, None)
        self.verify_imis_instance(imis_insuree)

    def test_create_object_from_json(self):
        self.setUp()
        fhir_patient = FHIRBaseObject.loads(self._test_patient_json_representation, 'json')
        self.verify_fhir_instance(fhir_patient)

    def test_fhir_object_to_json(self):
        self.setUp()
        fhir_patient = self.create_test_fhir_instance()
        actual_representation = fhir_patient.dumps(format_='json')
        self.assertEqual(self._test_patient_json_representation, actual_representation)



import logging

from policy.services import EligibilityRequest, EligibilityService, EligibilityResponse

from api_fhir_r4.converters import CoverageEligibilityRequestConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class CoverageEligibilityRequestSerializer(BaseFHIRSerializer):

    fhirConverter = CoverageEligibilityRequestConverter()
    logger = logging.getLogger(__name__)

    def create(self, validated_data):
        eligibility_request = EligibilityRequest(chf_id=validated_data.get('chf_id'),
                                                 service_code=validated_data.get('service_code'),
                                                 item_code=validated_data.get('item_code'))
        request = self.context.get("request")
        try:
            response = EligibilityService(request.user).request(eligibility_request)
        except TypeError:
            self.logger.warning('The insuree with chfid `{}` is not connected with policy. '
                                'The default eligibility response will be used.'
                                .format(validated_data.get('chfid')))
            response = self.create_default_eligibility_response()
        return response

    def create_default_eligibility_response(self):
        return EligibilityResponse(
            eligibility_request=None,
            prod_id=None,
            total_admissions_left=0,
            total_visits_left=0,
            total_consultations_left=0,
            total_surgeries_left=0,
            total_deliveries_left=0,
            total_antenatal_left=0,
            consultation_amount_left=0.0,
            surgery_amount_left=0.0,
            delivery_amount_left=0.0,
            hospitalization_amount_left=0.0,
            antenatal_amount_left=0.0,
            min_date_service=None,
            min_date_item=None,
            service_left=0,
            item_left=0,
            is_item_ok=False,
            is_service_ok=False
        )



from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.converters import LocationConverter
from api_fhir_r4.models import PractitionerRole
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin, FhirApiUpdateTestMixin, \
    FhirApiCreateTestMixin, LocationTestMixin, PractitionerTestMixin, FhirApiDeleteTestMixin
from api_fhir_r4.utils import TimeUtils


class PractitionerRoleAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, FhirApiCreateTestMixin,
                               FhirApiUpdateTestMixin, FhirApiDeleteTestMixin, APITestCase):

    base_url = '/api_fhir_r4/PractitionerRole/'
    _test_json_path = "/test/test_practitionerRole.json"
    _TEST_LOCATION_CODE = "12345678"
    _TEST_CLAIM_ADMIN_CODE = "1234abcd"
    _TEST_UPDATED_LOCATION_CODE = "newCode"
    _TEST_UPDATED_LOCATION_NAME = "newLocation"
    _TEST_LEGAL_FORM = "G"
    _TEST_ADMIN_USER_ID = 1

    def setUp(self):
        super(PractitionerRoleAPITests, self).setUp()

    def verify_updated_obj(self, updated_obj):
        self.assertTrue(isinstance(updated_obj, PractitionerRole))
        self.assertEqual(self._TEST_UPDATED_LOCATION_CODE,
                         LocationConverter.get_resource_id_from_reference(updated_obj.location[0]))
        self.assertEqual(self._TEST_UPDATED_LOCATION_NAME,
                         LocationConverter.get_imis_obj_by_fhir_reference(updated_obj.location[0]).name)

    def update_resource(self, data):
        new_location = self._create_and_save_hf()
        new_location.name = self._TEST_UPDATED_LOCATION_NAME
        new_location.code = self._TEST_UPDATED_LOCATION_CODE
        new_location.save()
        data['location'][0] = LocationConverter.build_fhir_resource_reference(new_location).toDict()

    def create_dependencies(self):
        self._create_and_save_hf()
        self._create_and_save_claim_admin()

    def _create_and_save_hf(self):
        imis_hf = LocationTestMixin().create_test_imis_instance()
        imis_hf.id = None
        imis_hf.validity_from = TimeUtils.now()
        imis_hf.offline = False
        imis_hf.audit_user_id = self._TEST_ADMIN_USER_ID
        imis_hf.code = self._TEST_LOCATION_CODE
        imis_hf.save()
        return imis_hf

    def _create_and_save_claim_admin(self):
        claim_admin = PractitionerTestMixin().create_test_imis_instance()
        claim_admin.id = None
        claim_admin.audit_user_id = self._TEST_ADMIN_USER_ID
        claim_admin.code = self._TEST_CLAIM_ADMIN_CODE
        claim_admin.save()
        return claim_admin

    def verify_deleted_response(self, response):
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        practitioner_role = self.get_fhir_obj_from_json_response(response)
        self.assertTrue(isinstance(practitioner_role, PractitionerRole))
        self.assertEqual(0, len(practitioner_role.location))




from claim.models import ClaimAdmin
from django.utils.translation import gettext

from api_fhir_r4.configurations import R4IdentifierConfig
from api_fhir_r4.converters import BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin
from fhir.resources.practitioner import Practitioner
from api_fhir_r4.utils import TimeUtils, DbManagerUtils


class PractitionerConverter(BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_claim_admin, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_practitioner = Practitioner.construct()
        cls.build_fhir_pk(fhir_practitioner, imis_claim_admin, reference_type)
        cls.build_fhir_identifiers(fhir_practitioner, imis_claim_admin)
        cls.build_human_names(fhir_practitioner, imis_claim_admin)
        cls.build_fhir_birth_date(fhir_practitioner, imis_claim_admin)
        cls.build_fhir_telecom(fhir_practitioner, imis_claim_admin)
        return fhir_practitioner

    @classmethod
    def to_imis_obj(cls, fhir_practitioner, audit_user_id):
        errors = []
        fhir_practitioner = Practitioner(**fhir_practitioner)
        imis_claim_admin = PractitionerConverter.create_default_claim_admin(audit_user_id)
        cls.build_imis_identifiers(imis_claim_admin, fhir_practitioner, errors)
        cls.build_imis_names(imis_claim_admin, fhir_practitioner)
        cls.build_imis_birth_date(imis_claim_admin, fhir_practitioner)
        cls.build_imis_contacts(imis_claim_admin, fhir_practitioner)
        cls.check_errors(errors)
        return imis_claim_admin

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_claim_admin_code_type()

    @classmethod
    def get_reference_obj_uuid(cls, claim_admin: ClaimAdmin):
        return claim_admin.uuid

    @classmethod
    def get_reference_obj_id(cls, claim_admin: ClaimAdmin):
        return claim_admin.id

    @classmethod
    def get_reference_obj_code(cls, claim_admin: ClaimAdmin):
        return claim_admin.code

    @classmethod
    def get_fhir_resource_type(cls):
        return Practitioner

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_claim_admin_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(ClaimAdmin, uuid=imis_claim_admin_uuid)

    @classmethod
    def create_default_claim_admin(cls, audit_user_id):
        imis_claim_admin = ClaimAdmin()
        imis_claim_admin.validity_from = TimeUtils.now()
        imis_claim_admin.audit_user_id = audit_user_id
        return imis_claim_admin

    @classmethod
    def build_fhir_identifiers(cls, fhir_practitioner, imis_claim_admin):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_claim_admin)
        fhir_practitioner.identifier = identifiers

    @classmethod
    def build_fhir_code_identifier(cls, identifiers, imis_claim_admin):
        if imis_claim_admin.code:
            identifier = cls.build_fhir_identifier(imis_claim_admin.code,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_claim_admin_code_type())
            identifiers.append(identifier)

    @classmethod
    def build_imis_identifiers(cls, imis_claim_admin, fhir_practitioner, errors):
        value = cls.get_fhir_identifier_by_code(fhir_practitioner.identifier,
                                                R4IdentifierConfig.get_fhir_claim_admin_code_type())
        if value:
            imis_claim_admin.code = value
        cls.valid_condition(imis_claim_admin.code is None, gettext('Missing the claim admin code'), errors)

    @classmethod
    def build_human_names(cls, fhir_practitioner, imis_claim_admin):
        name = cls.build_fhir_names_for_person(imis_claim_admin)
        fhir_practitioner.name = [name]

    @classmethod
    def build_imis_names(cls, imis_claim_admin, fhir_practitioner):
        names = fhir_practitioner.name
        imis_claim_admin.last_name, imis_claim_admin.other_names = cls.build_imis_last_and_other_name(names)

    @classmethod
    def build_fhir_birth_date(cls, fhir_practitioner, imis_claim_admin):
        if imis_claim_admin.dob is not None:
            from core import datetime
            # check if datetime object
            if isinstance(imis_claim_admin.dob, datetime.datetime):
                fhir_practitioner.birthDate = imis_claim_admin.dob.date().isoformat()
            else:
                fhir_practitioner.birthDate = imis_claim_admin.dob.isoformat()

    @classmethod
    def build_imis_birth_date(cls, imis_claim_admin, fhir_practitioner):
        birth_date = fhir_practitioner.birthDate
        if birth_date:
            imis_claim_admin.dob = TimeUtils.str_to_date(birth_date)

    @classmethod
    def build_fhir_telecom(cls, fhir_practitioner, imis_claim_admin):
        fhir_practitioner.telecom = cls.build_fhir_telecom_for_person(phone=imis_claim_admin.phone,
                                                                      email=imis_claim_admin.email_id)

    @classmethod
    def build_imis_contacts(cls, imis_claim_admin, fhir_practitioner):
        imis_claim_admin.phone, imis_claim_admin.email_id = cls.build_imis_phone_num_and_email(fhir_practitioner.telecom)



import copy

from insuree.models import Insuree, Gender, Education, Profession,Family

from api_fhir_r4.converters import PatientConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers import BaseFHIRSerializer


class PatientSerializer(BaseFHIRSerializer):
    fhirConverter = PatientConverter()

    def create(self, validated_data):
        chf_id = validated_data.get('chf_id')
        if Insuree.objects.filter(chf_id=chf_id).count() > 0:
            raise FHIRException('Exists patient with following chfid `{}`'.format(chf_id))
        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        obj = Insuree.objects.create(**copied_data)
        if copied_data['head']:
            obj.family=Family.objects.create(head_insuree=obj,audit_user_id=copied_data['audit_user_id'])
        obj.save()
        return obj

    def update(self, instance, validated_data):
        instance.last_name = validated_data.get('last_name', instance.last_name)
        instance.other_names = validated_data.get('other_names', instance.other_names)
        instance.passport = validated_data.get('passport', instance.passport)
        instance.dob = validated_data.get('dob', instance.dob)
        gender_code = validated_data.get('gender_id', instance.gender.code)
        instance.gender = Gender.objects.get(pk=gender_code)
        instance.marital = validated_data.get('marital', instance.marital)
        instance.phone = validated_data.get('phone', instance.phone)
        instance.email = validated_data.get('email', instance.email)
        instance.current_address = validated_data.get('current_address', instance.current_address)
        instance.geolocation = validated_data.get('geolocation', instance.geolocation)
        instance.audit_user_id = self.get_audit_user_id()
        instance.save()
        return instance



from claim.apps import ClaimConfig
from insuree.apps import InsureeConfig
from location.apps import LocationConfig
from policy.apps import PolicyConfig
from policyholder.apps import PolicyholderConfig
from medical.apps import MedicalConfig
from rest_framework import exceptions
from rest_framework.permissions import DjangoModelPermissions


class FHIRApiPermissions(DjangoModelPermissions):
    permissions_get = []
    permissions_post = []
    permissions_put = []
    permissions_patch = []
    permissions_delete = []

    def __init__(self):
        self.perms_map['GET'] = self.permissions_get
        self.perms_map['POST'] = self.permissions_post
        self.perms_map['PUT'] = self.permissions_put
        self.perms_map['PATCH'] = self.permissions_patch
        self.perms_map['DELETE'] = self.permissions_delete

    def get_required_permissions(self, method, model_cls):
        if method not in self.perms_map:
            raise exceptions.MethodNotAllowed(method)

        return self.perms_map[method]


class FHIRApiClaimPermissions(FHIRApiPermissions):
    permissions_get = ClaimConfig.gql_query_claims_perms
    permissions_post = ClaimConfig.gql_mutation_create_claims_perms
    permissions_put = ClaimConfig.gql_mutation_update_claims_perms
    permissions_patch = ClaimConfig.gql_mutation_update_claims_perms
    permissions_delete = ClaimConfig.gql_mutation_delete_claims_perms


class FHIRApiCommunicationRequestPermissions(FHIRApiPermissions):
    permissions_get = ClaimConfig.gql_mutation_select_claim_feedback_perms
    permissions_post = ClaimConfig.gql_mutation_deliver_claim_feedback_perms
    permissions_put = ClaimConfig.gql_mutation_deliver_claim_feedback_perms
    permissions_patch = ClaimConfig.gql_mutation_deliver_claim_feedback_perms
    permissions_delete = ClaimConfig.gql_mutation_skip_claim_feedback_perms


class FHIRApiPractitionerPermissions(FHIRApiPermissions):
    permissions_get = ClaimConfig.gql_query_claim_admins_perms
    permissions_post = ClaimConfig.gql_query_claim_admins_perms
    permissions_put = ClaimConfig.gql_query_claim_admins_perms
    permissions_patch = ClaimConfig.gql_query_claim_admins_perms
    permissions_delete = ClaimConfig.gql_query_claim_admins_perms


class FHIRApiCoverageEligibilityRequestPermissions(FHIRApiPermissions):
    permissions_get = PolicyConfig.gql_query_eligibilities_perms
    permissions_post = []
    permissions_put = []
    permissions_patch = []
    permissions_delete = []


class FHIRApiCoverageRequestPermissions(FHIRApiPermissions):
    permissions_get = PolicyConfig.gql_query_policies_by_insuree_perms
    permissions_post = []
    permissions_put = []
    permissions_patch = []
    permissions_delete = []


class FHIRApiHFPermissions(FHIRApiPermissions):
    permissions_get = LocationConfig.gql_query_health_facilities_perms
    permissions_post = LocationConfig.gql_mutation_create_health_facilities_perms
    permissions_put = LocationConfig.gql_mutation_create_health_facilities_perms
    permissions_patch = LocationConfig.gql_mutation_create_health_facilities_perms
    permissions_delete = LocationConfig.gql_mutation_delete_health_facilities_perms


class FHIRApiInsureePermissions(FHIRApiPermissions):
    permissions_get = InsureeConfig.gql_query_insurees_perms
    permissions_post = InsureeConfig.gql_mutation_create_insurees_perms
    permissions_put = InsureeConfig.gql_mutation_update_insurees_perms
    permissions_patch = InsureeConfig.gql_mutation_update_insurees_perms
    permissions_delete = InsureeConfig.gql_mutation_delete_insurees_perms


class FHIRApiMedicationPermissions(FHIRApiPermissions):
    permissions_get = MedicalConfig.gql_query_medical_items_perms
    permissions_post = MedicalConfig.gql_mutation_medical_items_add_perms
    permissions_put = MedicalConfig.gql_mutation_medical_items_update_perms
    permissions_patch = MedicalConfig.gql_mutation_medical_items_update_perms
    permissions_delete = MedicalConfig.gql_mutation_medical_items_delete_perms


class FHIRApiConditionPermissions(FHIRApiPermissions):
    permissions_get = MedicalConfig.gql_query_diagnosis_perms
    permissions_post = []
    permissions_put = []
    permissions_patch = []
    permissions_delete = []


class FHIRApiActivityDefinitionPermissions(FHIRApiPermissions):
    permissions_get = MedicalConfig.gql_query_medical_services_perms
    permissions_post = MedicalConfig.gql_mutation_medical_services_add_perms
    permissions_put = MedicalConfig.gql_mutation_medical_services_update_perms
    permissions_patch = MedicalConfig.gql_mutation_medical_services_update_perms
    permissions_delete = MedicalConfig.gql_mutation_medical_services_delete_perms


class FHIRApiHealthServicePermissions(FHIRApiPermissions):
    permissions_get = LocationConfig.gql_query_health_facilities_perms
    permissions_post = LocationConfig.gql_mutation_create_health_facilities_perms
    permissions_put = LocationConfig.gql_mutation_edit_health_facilities_perms
    permissions_patch = LocationConfig.gql_mutation_edit_health_facilities_perms
    permissions_delete = LocationConfig.gql_mutation_delete_health_facilities_perms


class FHIRApiGroupPermissions(FHIRApiPermissions):
    permissions_get = InsureeConfig.gql_query_families_perms
    permissions_post = InsureeConfig.gql_mutation_create_families_perms
    permissions_put = InsureeConfig.gql_mutation_update_families_perms
    permissions_patch = InsureeConfig.gql_mutation_update_families_perms
    permissions_delete = InsureeConfig.gql_mutation_delete_families_perms


class FHIRApiOrganizationPermissions(FHIRApiPermissions):
    permissions_get = PolicyholderConfig.gql_query_policyholder_perms
    permissions_post = PolicyholderConfig.gql_mutation_create_policyholder_perms
    permissions_put = PolicyholderConfig.gql_mutation_update_policyholder_perms
    permissions_patch = PolicyholderConfig.gql_mutation_update_policyholder_perms
    permissions_delete = PolicyholderConfig.gql_mutation_delete_policyholder_perms



from django.utils.translation import gettext
from api_fhir_r4.configurations import R4CoverageConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.models import ContractSignerV2 as ContractSigner
from fhir.resources.contract import Contract, ContractTermAssetContext, ContractTermAssetValuedItem, \
    ContractTerm, ContractTermAsset
from fhir.resources.money import Money
from fhir.resources.period import Period

from product.models import Product
from policy.models import Policy
from insuree.models import Insuree
from insuree.models import Family
from core.models import Officer
from api_fhir_r4.utils import DbManagerUtils,TimeUtils


class ContractConverter(BaseFHIRConverter, ReferenceConverterMixin):
    @classmethod
    def to_fhir_obj(cls, imis_policy, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_contract = Contract.construct()
        cls.build_contract_identifier(fhir_contract, imis_policy)
        contractTerm = ContractTerm.construct()
        contractTermAsset = ContractTermAsset.construct()
        cls.build_contract_asset_context(contractTermAsset, imis_policy, reference_type)
        cls.build_contract_asset_type_reference(contractTermAsset, imis_policy, reference_type)
        cls.build_contract_valued_item_entity(contractTermAsset, imis_policy)
        cls.build_contract_asset_use_period(contractTermAsset, imis_policy)
        contractTerm.asset = [contractTermAsset]
        # TODO - fix commented entries: 'term' and 'signer'
        #fhir_contract.term = [contractTerm]
        cls.build_contract_status(fhir_contract, imis_policy)
        #cls.build_contract_signer(fhir_contract, imis_policy, reference_type)
        cls.build_contract_state(fhir_contract, imis_policy)
        return fhir_contract
    
    @classmethod
    def to_imis_obj(cls,fhir_contract, audit_user_id):
        errors = []
        fhir_contract = Contract(**fhir_contract)
        imis_policy = Policy()
        imis_policy.audit_user_id = audit_user_id
        cls.build_imis_period(imis_policy, fhir_contract.term, errors)
        cls.build_imis_useperiod(imis_policy, fhir_contract.term, errors)
        cls.build_imis_status(fhir_contract,imis_policy, errors)
        cls.build_imis_signer(fhir_contract, imis_policy, errors)
        cls.build_imis_product(fhir_contract, imis_policy, errors)
        cls.build_imis_state(fhir_contract, imis_policy, errors)
        cls.build_imis_insurees(fhir_contract, imis_policy, errors)
        cls.check_errors(errors)
        return imis_policy
    @classmethod
    def get_reference_obj_uuid(cls, imis_policy: Policy):
        return imis_policy.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_policy: Policy):
        return imis_policy.id


    @classmethod
    def get_fhir_resource_type(cls):
        return Contract

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_policy_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Policy, code=imis_policy_code)

    @classmethod
    def build_contract_identifier(cls, fhir_contract, imis_policy):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_policy)
        fhir_contract.identifier = identifiers
        return fhir_contract

    @classmethod
    def build_all_identifiers(cls, identifiers, imis_object):
        # Coverage have only uuid coverage
        cls.build_fhir_uuid_identifier(identifiers, imis_object)
        cls.build_fhir_id_identifier(identifiers, imis_object)
        return identifiers

    @classmethod
    def build_contract_asset_use_period(cls, contract_asset, imis_policy):
        period_use = Period.construct()
        period= Period.construct()
        if imis_policy.start_date is not None:
            period.start = imis_policy.start_date.strftime("%Y-%m-%d")
            period_use.start = period.start
        if imis_policy.effective_date is not None:
            period_use.start = imis_policy.effective_date.strftime("%Y-%m-%d")
            if period_use.start is None:
                period.start = period_use.start
        if imis_policy.expiry_date is not None:
            period_use.end = imis_policy.expiry_date.strftime("%Y-%m-%d")
            period.end = period_use.end

        if type(contract_asset.usePeriod) is not list:
            contract_asset.usePeriod = [period_use]
        else:
            contract_asset.usePeriod.append(period_use)
        if type(contract_asset.period) is not list:
            contract_asset.period = [period]
        else:
            contract_asset.period.append(period)
        return contract_asset

    @classmethod
    def build_contract_asset_context(cls, contract_term_asset, imis_policy, reference_type):
        insureePolicies = imis_policy.insuree_policies.all()
        for insureePolicy in insureePolicies:
            if insureePolicy.insuree.head is True:
                party_role = cls.build_simple_codeable_concept(R4CoverageConfig.get_offer_insuree_role_code())
            else:
                party_role = cls.build_simple_codeable_concept(R4CoverageConfig.get_offer_dependant_role_code())

            assetContext = ContractTermAssetContext()
            assetContext.code = [party_role]

            if imis_policy.family.location:
                display = insureePolicy.insuree.uuid + ":" + imis_policy.family.location.code # used for the DHIS integration
            else:
                display = insureePolicy.insuree.uuid

            assetContext.reference = cls.build_fhir_resource_reference(
                insureePolicy.insuree,
                "Patient",
                display,
                reference_type=reference_type
            )

            if contract_term_asset.context is None:
                contract_term_asset.context = [assetContext]
            else:
                contract_term_asset.context.append(assetContext)
        return contract_term_asset
        
    @classmethod
    def build_contract_status(cls, contract, imis_policy):
        if imis_policy.status is imis_policy.STATUS_ACTIVE:
            contract.status = R4CoverageConfig.get_status_policy_code()
        elif imis_policy.status is imis_policy.STATUS_IDLE:
            contract.status = R4CoverageConfig.get_status_offered_code()
        elif imis_policy.status is imis_policy.STATUS_EXPIRED:
            contract.status = R4CoverageConfig.get_status_terminated_code()
        elif imis_policy.status is imis_policy.STATUS_SUSPENDED:
            contract.status = R4CoverageConfig.get_status_disputed_code()
        else:
            contract.status = imis_policy.status
        return contract

    @classmethod
    def build_contract_state(cls, contract, imis_policy):
        if imis_policy.stage is imis_policy.STAGE_NEW:
            contract.legalState = cls.build_simple_codeable_concept(R4CoverageConfig.get_status_offered_code())
        elif imis_policy.stage is imis_policy.STAGE_RENEWED:
            contract.legalState = cls.build_simple_codeable_concept(R4CoverageConfig.get_status_renewed_code())
        else:
            contract.legalState = cls.build_simple_codeable_concept(imis_policy.stage)
        return contract

    @classmethod
    def build_contract_valued_item_entity(cls, contract_asset, imis_policy):
        valued_item = ContractTermAssetValuedItem.construct()
        typeReference = cls.build_fhir_resource_reference(imis_policy.product, "InsuranceProduct", imis_policy.product.code )
        valued_item.entityReference = typeReference
        policy_value = Money.construct()
        policy_value.value = imis_policy.value
        valued_item.net = policy_value
        if type(contract_asset.valuedItem) is not list:
            contract_asset.valuedItem = [valued_item]
        else:
            contract_asset.valuedItem.append(valued_item)
        return contract_asset

    @classmethod
    def build_contract_asset_type_reference(cls, contract_asset, imis_policy, reference_type):
        typeReference = cls.build_fhir_resource_reference(
            imis_policy.product, "InsurancePlan", imis_policy.product.code, reference_type=reference_type)
        contract_asset.typeReference = [typeReference]
        return contract_asset

    @classmethod
    def build_contract_signer(cls, contract, imis_policy, reference_type):
        if imis_policy.officer is not None:
            reference = cls.build_fhir_resource_reference(
                imis_policy.officer, "Practitioner", reference_type=reference_type)
            signer = ContractSigner.construct()
            signer.party = reference
            eo_codeable_concept_type = cls.build_simple_codeable_concept(R4CoverageConfig.get_signer_eo_type_code())
            eo_type = cls.get_first_coding_from_codeable_concept(eo_codeable_concept_type)
            signer.type = eo_type
            if contract.signer is None:
                contract.signer = [signer]
            else:
                if type(contract.signer) is not list:
                    contract.signer = [signer]
                else:
                    contract.signer.append(signer)
        if imis_policy.family is not None:
            if imis_policy.family.head_insuree is not None:
                reference = cls.build_fhir_resource_reference(
                    imis_policy.family.head_insuree, "Patient", reference_type=reference_type)
                signer = ContractSigner.construct()
                signer.party = reference
                eo_codeable_concept_type = cls.build_simple_codeable_concept(R4CoverageConfig.get_signer_head_type_code())
                eo_type = cls.get_first_coding_from_codeable_concept(eo_codeable_concept_type)
                signer.type = eo_type
                if contract.signer is None:
                    contract.signer = [signer]
                else:
                    if type(contract.signer) is not list:
                        contract.signer = [signer]
                    else:
                        contract.signer.append(signer)

    @classmethod
    def build_imis_period(cls, imis_policy,fhir_contract,errors):
        for term in  fhir_contract:
            if term.asset:
                for asset in term.asset:
                    if asset.period:
                        for period in asset.period:
                            if not cls.valid_condition(period.start is None, gettext('Missing  `period start` attribute'),errors):
                                imis_policy.start_date = TimeUtils.str_to_date(period.start) 
                                imis_policy.enroll_date = TimeUtils.str_to_date(period.start) 
                            if not cls.valid_condition(period.end is None, gettext('Missing  `period end` attribute'),errors):
                                imis_policy.expiry_date = TimeUtils.str_to_date(period.end) 
                                imis_policy.validity_to = TimeUtils.str_to_date(period.end)
                    else:
                        cls.valid_condition(not asset.period, gettext('Missing  `period` attribute'),errors)
                        
                        
    @classmethod
    def build_imis_useperiod(cls, imis_policy,fhir_contract,errors):
        for term in  fhir_contract:
            if term.asset:
                for asset in term.asset:
                    if asset.usePeriod:
                        for period in asset.usePeriod:
                            if not cls.valid_condition(period.start is None, gettext('Missing  `usePeriod start` attribute'),errors):
                                imis_policy.effective_date = TimeUtils.str_to_date(period.start)
                            if not cls.valid_condition(period.end is None, gettext('Missing  `usePeriod end` attribute'),errors):
                                imis_policy.expiry_date = TimeUtils.str_to_date(period.end) 
                    else:
                        cls.valid_condition(not asset.usePeriod, gettext('Missing  `usePeriod` attribute'),errors)
    
    @classmethod
    def build_imis_status(cls, fhir_contract, imis_policy,errors):
        if fhir_contract.status:
            if fhir_contract.status == R4CoverageConfig.get_status_idle_code():
                imis_policy.status = cls.imis_map_status(R4CoverageConfig.get_status_idle_code(),imis_policy)
            elif fhir_contract.status == R4CoverageConfig.get_status_active_code():
                imis_policy.status = cls.imis_map_status(R4CoverageConfig.get_status_active_code(),imis_policy)
            elif fhir_contract.status == R4CoverageConfig.get_status_suspended_code():
                 imis_policy.status = cls.imis_map_status(R4CoverageConfig.get_status_suspended_code(),imis_policy)
            elif fhir_contract.status == R4CoverageConfig.get_status_expired_code():
                 imis_policy.status = cls.imis_map_status(R4CoverageConfig.get_status_expired_code(),imis_policy)
            else:
                pass    
        else:
            cls.valid_condition(fhir_contract.status is None, gettext('Missing  `status` attribute'),errors)

    @classmethod
    def imis_map_status(cls,code,imis_policy):
        status={
            R4CoverageConfig.get_status_idle_code():imis_policy.STATUS_IDLE,
            R4CoverageConfig.get_status_active_code():imis_policy.STATUS_ACTIVE,
            R4CoverageConfig.get_status_suspended_code():imis_policy.STATUS_SUSPENDED,
            R4CoverageConfig.get_status_expired_code():imis_policy.STATUS_EXPIRED,
        }
        return status[code]

    @classmethod
    def build_imis_signer(cls,fhir_contract, imis_policy,errors):
        if fhir_contract.signer:
            for signer in  fhir_contract.signer:
                if signer.type:
                    if signer.type.text and signer.party.reference is not None:
                        if signer.type.text =='HeadOfFamily':
                            reference = signer.party.reference.split("/", 2)
                            
                            try:
                                insuree = Insuree.objects.get(uuid=reference[1])
                                if insuree.head:
                                    imis_policy.family= Family.objects.filter(head_insuree=insuree).first()
                                else:
                                    cls.valid_condition(True, gettext('Missing  `Member details provided belong to a depedant` attribute'),errors)
                            except:
                                cls.valid_condition(True, gettext('Missing  `Family head provided does not exist` attribute'),errors)
                        elif signer.type.text == 'EnrolmentOfficer':
                            reference = signer.party.reference.split("/", 2)
                            imis_policy.officer = Officer.objects.get(uuid=reference[1])
                        else:
                            pass     
                else:
                    cls.valid_condition(signer.type is None, gettext('Missing  `type` attribute'),errors)
        else:
            cls.valid_condition(not fhir_contract.signer, gettext('Missing  `signer` attribute'),errors)
            
    @classmethod
    def build_imis_insurees(cls,fhir_contract,imis_policy,errors):
        if fhir_contract.term:
            insurees =[]
            for term in  fhir_contract.term:
                if term.asset:
                    for asset in term.asset:
                        if asset.context:
                            for item in asset.context:
                               if item.reference is not None:
                                   reference = item.reference.reference.split("/", 2)
                                   obj = Insuree.objects.get(uuid=reference[1])
                                   if imis_policy.family_id is not None:
                                       if obj.family == imis_policy.family:
                                           if type(insurees) is not list:
                                               insurees = [obj.uuid]
                                           else:
                                               insurees.append(obj.uuid)
                                       else:
                                            if 'Missing  `Invalid Context reference` attribute' not in errors:
                                                cls.valid_condition(True, gettext('Missing  `Invalid Context reference` attribute'),errors)
                            imis_policy.insurees = insurees                    
                        else:
                            cls.valid_condition(not asset.context, gettext('Missing  `context` attribute'),errors)
                else:
                    cls.valid_condition(not term.asset, gettext('Missing  `asset` attribute'),errors)
                        
        else:
            cls.valid_condition(not fhir_contract, gettext('Missing  `term` attribute'),errors)
            
    
    
    @classmethod
    def build_imis_product(cls,fhir_contract,imis_policy,errors):
        if fhir_contract.term:
            for term in  fhir_contract.term:
                if term.asset:
                    for asset in term.asset:
                        if asset.valuedItem:
                            for item in asset.valuedItem:
                                if item.entityReference is not None:
                                    if item.entityReference.reference is not None:
                                        reference = item.entityReference.reference.split("/", 2)
                                        imis_policy.product = Product.objects.get(uuid=reference[1])
                                if item.net is not None:
                                    if item.net.value is not None:
                                        imis_policy.value = item.net.value                                                       
                        else:
                            cls.valid_condition(not asset.valuedItem, gettext('Missing  `valuedItem` attribute'),errors)
                else:
                    cls.valid_condition(not term.asset, gettext('Missing  `asset` attribute'),errors)
                        
        else:
            cls.valid_condition(not fhir_contract, gettext('Missing  `term` attribute'),errors)
                        
    @classmethod
    def build_imis_state(cls,fhir_contract, imis_policy,errors):
        if fhir_contract.legalState:
            if fhir_contract.legalState.text:
                if fhir_contract.legalState.text == R4CoverageConfig.get_status_offered_code():
                    imis_policy.stage = cls.imis_map_stage(R4CoverageConfig.get_status_offered_code(),imis_policy)
                elif fhir_contract.legalState.text == R4CoverageConfig.get_status_renewed_code():
                    imis_policy.stage = cls.imis_map_stage(R4CoverageConfig.get_status_renewed_code(),imis_policy)
                else:
                    pass       
        else:
            cls.valid_condition(fhir_contract.legalState is None, gettext('Missing  `legalState` attribute'),errors)
    
    @classmethod
    def imis_map_stage(cls,code,imis_policy):
        codes = {
            R4CoverageConfig.get_status_offered_code():imis_policy.STAGE_NEW,
            R4CoverageConfig.get_status_active_code():imis_policy.STAGE_RENEWED
        }
        return codes[code]



from unittest.mock import patch, MagicMock

from django.test import TestCase

from api_fhir_r4.mixins import ContainedContentSerializerMixin
from api_fhir_r4.models import FHIRBaseObject


class ContainedContentHelper(object):
    @staticmethod
    def build_test_converter(returned_obj=FHIRBaseObject()):
        converter = MagicMock()
        converter.convert_from_source = MagicMock(name='convert_from_source', return_value=[returned_obj])
        return converter


class ContainedContentSerializerMixinTestCase(TestCase):

    class BaseTestSerializer:
        context = {'contained': True}

        def to_representation(self, obj):
            return FHIRBaseObject().toDict()

    class TestSerializer(ContainedContentSerializerMixin, BaseTestSerializer):

        @property
        def contained_resources(self):
            return [
                ContainedContentHelper.build_test_converter()
            ]

    def test_resource_transformation(self):
        test_serializer = self.TestSerializer()
        test_imis_obj = MagicMock()
        representation = test_serializer.to_representation(test_imis_obj)

        expected_outcome = {'contained': [{}]}
        self.assertEqual(representation, expected_outcome)



from django.utils.translation import gettext
from policyholder.models import PolicyHolder
from location.models import Location
from api_fhir_r4.configurations import R4IdentifierConfig, GeneralConfiguration
from api_fhir_r4.converters import BaseFHIRConverter,ReferenceConverterMixin
from api_fhir_r4.converters.healthcareServiceConverter import HealthcareServiceConverter
from api_fhir_r4.converters.locationConverter import LocationConverter
from fhir.resources.extension import Extension
from fhir.resources.attachment import Attachment
from fhir.resources.coding import Coding
from fhir.resources.contactpoint import ContactPoint
from fhir.resources.organization import Organization, OrganizationContact
from api_fhir_r4.utils import TimeUtils, DbManagerUtils


class OrganisationConverter(BaseFHIRConverter):
    
    @classmethod
    def to_fhir_obj(cls, imis_organisation, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        # TODO reshape this fhir object and imis, fix BankAccount
        fhir_organisation = Organization()
        cls.build_fhir_pk(fhir_organisation, imis_organisation.id)
        # TODO - locaction - as an extension?
        #cls.build_fhir_location(fhir_organisation, imis_organisation)
        cls.build_fhir_name(fhir_organisation, imis_organisation)
        # TODO - fix contact name field
        #cls.build_fhir_contact_name(imis_organisation, fhir_organisation)
        cls.build_fhir_date(fhir_organisation, imis_organisation)
        cls.build_fhir_legal_form(fhir_organisation, imis_organisation)
        cls.build_fhir_phone(fhir_organisation, imis_organisation)
        cls.build_fhir_fax(fhir_organisation, imis_organisation)
        cls.build_fhir_email(fhir_organisation, imis_organisation)
        cls.build_fhir_code(fhir_organisation, imis_organisation)
        cls.build_fhir_addresses(fhir_organisation, imis_organisation)
        cls.build_fhir_accountancy_account(fhir_organisation, imis_organisation)
        #cls.build_fhir_bank_account(fhir_organisation,imis_organisation)
        return fhir_organisation

    @classmethod
    def to_imis_obj(cls, fhir_organisation, audit_user_id):
        errors = []
        fhir_organisation = Organization(**fhir_organisation)
        imis_organisation = PolicyHolder()
        imis_organisation.date_created = TimeUtils.now()
        imis_organisation.date_updated = TimeUtils.now()
        cls.build_imis_addresses(imis_organisation,fhir_organisation)
        cls.build_imis_identifiers(imis_organisation,fhir_organisation)
        cls.build_imis_name(imis_organisation,fhir_organisation)
        cls.build_imis_code(imis_organisation,fhir_organisation,errors)
        cls.build_imis_phone(imis_organisation,fhir_organisation,errors)
        cls.build_imis_email(imis_organisation,fhir_organisation)
        cls.build_imis_contact(imis_organisation,fhir_organisation)
        cls.build_imis_fax(imis_organisation,fhir_organisation)
        cls.build_imis_bank_account(imis_organisation,fhir_organisation)
        cls.build_imis_accountancy_account(imis_organisation,fhir_organisation)
        cls.build_imis_legal_form(imis_organisation,fhir_organisation,errors)
        cls.check_errors(errors)
        return imis_organisation

    @classmethod
    def build_imis_addresses(cls,imis_organisation,fhir_organisation):
        addresses = fhir_organisation.address
        address={}
        if addresses is not None:
            address['text']=addresses[0].text
            address['type']=addresses[0].type
            address['use']=addresses[0].use
        imis_organisation.address=address

    @classmethod
    def get_reference_obj_id(cls, imis_organisation):
        return imis_organisation.uuid

    @classmethod
    def get_fhir_resource_type(cls):
        return Organization

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_organisation_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Insuree, uuid=imis_organisation_uuid)

    @classmethod
    def build_imis_code(cls,imis_organisation,fhir_organisation,errors):
        code =fhir_organisation.code
        if not cls.valid_condition(code is None,gettext('Missing  `code` attribute'), errors):
            imis_organisation.code=fhir_organisation.code
         
    @classmethod
    def build_imis_phone(cls,imis_organisation,fhir_organisation,errors):
        phone=fhir_organisation.phone
        if not cls.valid_condition(phone is None,gettext('Missing  `phone` attribute'), errors):
            imis_organisation.phone=fhir_organisation.phone
            
    @classmethod
    def build_imis_legal_form(cls,imis_organisation,fhir_organisation,errors):
        legal_form=fhir_organisation.legal_form
        if not cls.valid_condition(legal_form is None,gettext('Missing  `legal_form` attribute'), errors):
            imis_organisation.legal_form = fhir_organisation.legal_form
    
    @classmethod
    def build_imis_fax(cls,imis_organisation,fhir_organisation):
        imis_organisation.fax = fhir_organisation.fax 
    
    @classmethod
    def build_imis_bank_account(cls,imis_organisation,fhir_organisation):
        bank = cls.build_bank(fhir_organisation.bank_account)
        imis_organisation.bank_account = bank
    
    @classmethod
    def build_imis_accountancy_account(cls,imis_organisation,fhir_organisation):
        if fhir_organisation.accountancy_account:
        #    print(fhir_organisation.accountancy_account)
           imis_organisation.accountancy_account = fhir_organisation.accountancy_account
    
    @classmethod
    def build_imis_contact(cls,imis_organisation,fhir_organisation):
        if fhir_organisation.contact:
            imis_organisation.contact_name = fhir_organisation.contact[0]
        
    @classmethod
    def build_imis_email(cls,imis_organisation,fhir_organisation):
        imis_organisation.email=fhir_organisation.email 
      
    @classmethod
    def build_fhir_name(cls,fhir_organisation, imis_organisation):
        name = imis_organisation.trade_name
        fhir_organisation.name = name
        
        
        
    @classmethod
    def build_imis_name(cls,imis_organisation,fhir_organisation):
        imis_organisation.trade_name = fhir_organisation.name
         
    @classmethod
    def build_fhir_phone(cls,fhir_organisation, imis_organisation):
        phone = imis_organisation.phone
        fhir_organisation.phone = phone
        
    @classmethod
    def build_fhir_legal_form(cls,fhir_organisation, imis_organisation):
        fhir_organisation.legal_form = imis_organisation.legal_form
    @classmethod
    def build_fhir_date(cls,fhir_organisation, imis_organisation):
        fhir_organisation.date_created = imis_organisation.date_created.isoformat()
        
    @classmethod
    def build_fhir_phone(cls,fhir_organisation, imis_organisation):
        phone = imis_organisation.phone
        fhir_organisation.phone = phone
    
    @classmethod
    def build_fhir_email(cls,fhir_organisation, imis_organisation):
        email = imis_organisation.email
        fhir_organisation.email = email
        
    @classmethod
    def build_fhir_code(cls,fhir_organisation, imis_organisation):
        code = imis_organisation.code
        fhir_organisation.code = code
    
    @classmethod
    def build_fhir_accountancy_account(cls,fhir_organisation, imis_organisation):
        if imis_organisation.accountancy_account:
           fhir_organisation.accountancy_account = imis_organisation.accountancy_account
    @classmethod
    def build_fhir_bank_account(cls,fhir_organisation, imis_organisation):
        if imis_organisation.bank_account:
           fhir_organisation.bank_account = imis_organisation.bank_account
    @classmethod
    def build_fhir_fax(cls,fhir_organisation, imis_organisation):
        fax = imis_organisation.fax
        fhir_organisation.fax = fax
        
    @classmethod
    def build_fhir_identifiers(cls, fhir_organisation, imis_organisation):
        identifiers = []
        cls.build_fhir_uuid_identifier(identifiers, imis_organisation)
        fhir_organisation.identifier = identifiers

    @classmethod
    def build_imis_identifiers(cls, imis_organisation, fhir_organisation):
        value = cls.get_fhir_identifier_by_code(fhir_organisation.identifier,
                                                R4IdentifierConfig.get_fhir_uuid_type_code())
        # if value:
        #     imis_organisation.id= value
     
    @classmethod
    def build_fhir_location(cls, fhir_organisation, imis_organisation):
        locations = []
        if imis_organisation.locations is not None:
            locations.append({"name": imis_organisation.locations.name})
        if type(fhir_organisation.location) is not list:
            fhir_organisation.location = locations
        else:
            fhir_organisation.location.append(locations)
    

    @classmethod
    def build_fhir_telecom(cls, fhir_organisation, imis_organisation):
        fhir_organisation.telecom = cls.build_fhir_telecom_for_person(phone=imis_organisation.phone, email=imis_organisation.email)

    @classmethod
    def build_imis_contacts(cls, imis_organisation, fhir_organisation):
        imis_organisation.phone, imis_organisation.email = cls.build_imis_phone_num_and_email(fhir_organisation.telecom)

    @classmethod
    def build_fhir_addresses(cls, fhir_organisation, imis_organisation):
        addresses = []
        if imis_organisation.address is not None:
            addresses.append(imis_organisation.address)
        if type(fhir_organisation.address) is not list:
            fhir_organisation.address = addresses
        else:
            fhir_organisation.address.append(addresses)


    @classmethod
    def build_fhir_contact_name(cls,imis_organisation,fhir_organisation):
        contacts =[]
        #contact = ContactPoint()
        if imis_organisation.contact_name is not None:
            # contact.system = imis_organisation.contact_name['system']
            contacts.append(imis_organisation.contact_name)
        if type(fhir_organisation.contact) is not list:
            fhir_organisation.contact = contacts
        else:
            fhir_organisation.contact.append(contacts)
        


from api_fhir_r4.converters import OperationOutcomeConverter
from api_fhir_r4.permissions import FHIRApiClaimPermissions, FHIRApiCoverageEligibilityRequestPermissions, \
    FHIRApiCoverageRequestPermissions, FHIRApiCommunicationRequestPermissions, FHIRApiPractitionerPermissions, \
    FHIRApiHFPermissions, FHIRApiInsureePermissions, FHIRApiMedicationPermissions, FHIRApiConditionPermissions, \
    FHIRApiActivityDefinitionPermissions, FHIRApiHealthServicePermissions, FHIRApiGroupPermissions, \
    FHIRApiOrganizationPermissions
from claim.models import ClaimAdmin, Claim, Feedback, ClaimItem, ClaimService
from core.models import User
from django.db.models import OuterRef, Exists
from insuree.models import Insuree, InsureePolicy,Family
from policyholder.models import PolicyHolder
from location.models import HealthFacility, Location
from policy.models import Policy
from medical.models import Item, Diagnosis, Service
from rest_framework import exceptions, viewsets, mixins, status
from rest_framework.authentication import SessionAuthentication
from rest_framework.response import Response
from rest_framework.permissions import AllowAny
from rest_framework.views import APIView
from rest_framework.decorators import parser_classes,permission_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.viewsets import GenericViewSet
import os
from api_fhir_r4.paginations import FhirBundleResultsSetPagination
from api_fhir_r4.permissions import FHIRApiPermissions
from api_fhir_r4.configurations import R4CoverageEligibilityConfiguration as Config, GeneralConfiguration
from api_fhir_r4.serializers import PatientSerializer, LocationSerializer, LocationSiteSerializer, PractitionerRoleSerializer, \
    PractitionerSerializer, ClaimSerializer, CoverageEligibilityRequestSerializer, \
    PolicyCoverageEligibilityRequestSerializer, ClaimResponseSerializer, CommunicationRequestSerializer, \
    MedicationSerializer, ConditionSerializer, ActivityDefinitionSerializer, HealthcareServiceSerializer, ContractSerializer, GroupSerializer,\
    OrganisationSerializer
from api_fhir_r4.serializers.coverageSerializer import CoverageSerializer

from django.db.models import Q, Prefetch
from graphql_jwt.utils import jwt_payload
from core.jwt import *
from django.db.models import Prefetch
from rest_framework.decorators import api_view
import datetime
from rest_framework.parsers import JSONParser


class CsrfExemptSessionAuthentication(SessionAuthentication):
    def enforce_csrf(self, request):
        return


class LoginView(viewsets.ViewSet):
    permission_classes = (AllowAny,)

    def create(self, request, *args, **kwargs):
        data = request.data
        # check if we have both required data in request payload
        if 'username' in data and 'password' in data:
            username = data['username']
            password = data['password']
            users = User.objects.filter(username=username)
            # check if user with such username exists
            if len(list(users)) == 1:
                # validate provided password from payload
                user = users.first()
                if user.check_password(password):
                    # set the user to context
                    request.user = user
                    # take the payload base on user data - using same mechanism as
                    # in graphql_jwt with generating payload.
                    payload = jwt_payload(user=user)
                    # encode token based on payload
                    token = jwt_encode_user_key(payload=payload, context=request)
                    if token:
                        # return ok
                        response = {
                            "token": token,
                            "exp": payload['exp'],
                        }
                        return Response(data=response, status=200)
            # return unauthorized
            return Response(status=401)
        else:
            # return bad request
            return Response(status=400)


class BaseFHIRView(APIView):
    pagination_class = FhirBundleResultsSetPagination
    permission_classes = (FHIRApiPermissions,)
    authentication_classes = [CsrfExemptSessionAuthentication] + APIView.settings.DEFAULT_AUTHENTICATION_CLASSES


class InsureeViewSet(BaseFHIRView, viewsets.ModelViewSet):
    lookup_field = 'uuid'
    serializer_class = PatientSerializer
    permission_classes = (FHIRApiInsureePermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset().select_related('gender').select_related('photo').select_related('family__location')
        refDate = request.GET.get('refDate')
        claim_date = request.GET.get('claimDateFrom')
        identifier = request.GET.get("identifier")
        if identifier:
            queryset = queryset.filter(chf_id=identifier)
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')
            if refDate != None:
                # isValidDate = True
                try :
                    datevar = datetime.datetime.strptime(refDate, "%Y-%m-%d").date()
                except ValueError :
                    pass
                    # isValidDate = False
                #datevar = refDate
                queryset = queryset.filter(validity_from__gte=datevar)
            if claim_date is not None:
                #year,month,day = refDate.split('-')
                try:
                    claim_parse_dated = datetime.datetime.strptime(claim_date, "%Y-%m-%d").date()
                    #datetime.datetime(int(year), int(month), int(day))
                except ValueError:
                    result = OperationOutcomeConverter.build_for_400_bad_request("claimDateFrom should be in dd-mm-yyyy format")
                    return Response(result.dict(), status.HTTP_400_BAD_REQUEST)
                has_claim_in_range = Claim.objects\
                    .filter(date_claimed__gte=claim_parse_dated)\
                    .filter(insuree_id=OuterRef("id"))\
                    .values("id")
                queryset = queryset.annotate(has_claim_in_range=Exists(has_claim_in_range)).filter(has_claim_in_range=True)

        serializer = PatientSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def get_queryset(self):
        return Insuree.get_queryset(None, self.request.user)


class GroupViewSet(BaseFHIRView, viewsets.ModelViewSet):
    lookup_field = 'uuid'
    serializer_class = GroupSerializer
    permission_classes = (FHIRApiGroupPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        identifier = request.GET.get("identifier") 
        if identifier:
            queryset = queryset.filter(head_insuree_id__chf_id=identifier)
        else:
            queryset = queryset.filter(validity_to__isnull=True)
        serializer = GroupSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def get_queryset(self):
        return Family.objects.all().order_by('validity_from')


class OrganisationViewSet(BaseFHIRView, viewsets.ModelViewSet):
    lookup_field = 'id'
    serializer_class = OrganisationSerializer
    permission_classes = (FHIRApiOrganizationPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        identifier = request.GET.get("code")       
        if identifier:
            queryset = queryset.filter(code=identifier)
        else:
            queryset = queryset.filter().order_by('date_created')
        serializer = OrganisationSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def get_queryset(self):
        return PolicyHolder.objects.filter(is_deleted=False).order_by('date_created')


class LocationViewSet(BaseFHIRView, viewsets.ModelViewSet):
    lookup_field = 'uuid'
    serializer_class = LocationSerializer
    permission_classes = (FHIRApiHFPermissions,)

    def list(self, request, *args, **kwargs):
        identifier = request.GET.get("identifier")
        physicalType = request.GET.get('physicalType')
        queryset = self.get_queryset(physicalType)
        if identifier:
            queryset = queryset.filter(code=identifier)
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')
        if (physicalType and physicalType == 'si'):
            self.serializer_class=LocationSiteSerializer
            serializer = LocationSiteSerializer(self.paginate_queryset(queryset), many=True)
        else:
            serializer = LocationSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def retrieve(self, *args, **kwargs):
        physicalType = self.request.GET.get('physicalType')
        if ( physicalType and physicalType == 'si'):
            self.serializer_class=LocationSiteSerializer
            self.queryset = self.get_queryset('si')
        response = viewsets.ModelViewSet.retrieve(self, *args, **kwargs)
        return response

    def get_queryset(self, physicalType = 'area'):
        if physicalType == 'si':
            hf_queryset = HealthFacility.get_queryset(None, self.request.user)
            return hf_queryset.select_related('location').select_related('sub_level').select_related('legal_form')
        else:
            return Location.get_queryset(None, self.request.user)


class PractitionerRoleViewSet(BaseFHIRView, viewsets.ModelViewSet):
    lookup_field = 'uuid'
    serializer_class = PractitionerRoleSerializer
    permission_classes = (FHIRApiPractitionerPermissions,)

    def list(self, request, *args, **kwargs):
        identifier = request.GET.get("identifier")
        queryset = self.get_queryset()
        if identifier:
            queryset = queryset.filter(code=identifier)
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')

        serializer = PractitionerRoleSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def perform_destroy(self, instance):
        instance.health_facility_id = None
        instance.save()

    def get_queryset(self):
        return ClaimAdmin.get_queryset(None, self.request.user)


class PractitionerViewSet(BaseFHIRView, viewsets.ModelViewSet):
    lookup_field = 'uuid'
    serializer_class = PractitionerSerializer
    permission_classes = (FHIRApiPractitionerPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        identifier = request.GET.get("identifier")
        if identifier:
            queryset = queryset.filter(code=identifier)
        serializer = PractitionerSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def get_queryset(self):
        return ClaimAdmin.get_queryset(None, self.request.user)


class ClaimViewSet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin,
                   mixins.CreateModelMixin, GenericViewSet):
    lookup_field = 'uuid'
    serializer_class = ClaimSerializer
    permission_classes = (FHIRApiClaimPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset().select_related('insuree').select_related('health_facility').select_related('icd')\
            .select_related('icd_1').select_related('icd_2').select_related('icd_3').select_related('icd_4')\
            .prefetch_related(Prefetch('items', queryset=ClaimItem.objects.filter(validity_to__isnull=True)))\
            .prefetch_related(Prefetch('services', queryset=ClaimService.objects.filter(validity_to__isnull=True)))\
            .prefetch_related(Prefetch('insuree__insuree_policies', queryset=InsureePolicy.objects.filter(validity_to__isnull=True).select_related("policy"))) #.filter(start_date__lte=date_from , expiry_date__gte=date_from)
        refDate = request.GET.get('refDate')
        identifier = request.GET.get("identifier")
        patient = request.GET.get("patient")
        contained = bool(request.GET.get("contained"))

        if identifier is not None:
            queryset = queryset.filter(Q(code=identifier) | Q(id=identifier) | Q(uuid=identifier))
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')
            if refDate is not None:
                year,month,day = refDate.split('-')
                # isValidDate = True
                try :
                    datetime.datetime(int(year),int(
                        
                        month),int(day))
                except ValueError :
                    pass
                    #isValidDate = False
                datevar = refDate
                queryset = queryset.filter(validity_from__gte=datevar)
            if patient is not None:
                for_patient = Insuree.objects\
                    .filter(indentifier = patient)\
                    .values("id")
                queryset = queryset.filter(chf_id in indentifier.id)
        serializer = ClaimSerializer(self.paginate_queryset(queryset), many=True, context={'contained': contained})
        return self.get_paginated_response(serializer.data)

    def retrieve(self, request, *args, **kwargs):
        contained = bool(request.GET.get("contained"))
        instance = self.get_object()
        serializer = self.get_serializer(instance, context={'contained': contained})
        return Response(serializer.data)

    def get_queryset(self):
        return Claim.get_queryset(None, self.request.user)


class ClaimResponseViewSet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin, GenericViewSet,
                           mixins.UpdateModelMixin):
    lookup_field = 'uuid'
    serializer_class = ClaimResponseSerializer
    permission_classes = (FHIRApiClaimPermissions,)

    def get_queryset(self):
        return Claim.get_queryset(None, self.request.user)


class CommunicationRequestViewSet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin, GenericViewSet):
    lookup_field = 'uuid'
    serializer_class = CommunicationRequestSerializer
    permission_classes = (FHIRApiCommunicationRequestPermissions,)

    def get_queryset(self):
        return Feedback.get_queryset(None, self.request.user)


class CoverageEligibilityRequestViewSet(BaseFHIRView, mixins.CreateModelMixin, GenericViewSet):
    queryset = Insuree.filter_queryset()
    serializer_class = eval(Config.get_serializer())
    permission_classes = (FHIRApiCoverageEligibilityRequestPermissions,)

    def get_queryset(self):
        return Insuree.get_queryset(None, self.request.user)
    
  
class CoverageRequestQuerySet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin,mixins.UpdateModelMixin,mixins.CreateModelMixin,GenericViewSet):
    lookup_field = 'uuid'
    serializer_class = CoverageSerializer
    permission_classes = (FHIRApiCoverageRequestPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset() 
        queryset.prefetch_related('services')
        refDate = request.GET.get('refDate')
        refEndDate = request.GET.get('refEndDate')
        identifier = request.GET.get("identifier")
        if identifier:
            queryset = queryset.filter(chf_id=identifier)
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')
            if refDate != None:
                isValidDate = True
                try :
                    datevar = datetime.datetime.strptime(refDate, "%Y-%m-%d").date()
                except ValueError :
                    isValidDate = False
                queryset = queryset.filter(validity_from__gte=datevar)
            if refEndDate != None:
                isValidDate = True
                try :
                    datevar = datetime.datetime.strptime(refEndDate, "%Y-%m-%d").date()
                except ValueError :
                    isValidDate = False
                queryset = queryset.filter(validity_from__lt=datevar)

        serializer = CoverageSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def get_queryset(self):
        return Policy.get_queryset(None, self.request.user)


class ContractViewSet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin, mixins.CreateModelMixin, GenericViewSet):
    lookup_field = 'uuid'
    serializer_class = ContractSerializer
    permission_classes = (FHIRApiCoverageRequestPermissions,)

    def list(self, request, *args, **kwargs):
        #queryset = self.get_queryset() 
        queryset =   self.get_queryset().select_related('product').select_related('officer')\
            .select_related('family__head_insuree').select_related('family__location')\
            .prefetch_related(Prefetch('insuree_policies', queryset=InsureePolicy.objects.filter(validity_to__isnull=True).select_related('insuree')))
        refDate = request.GET.get('refDate')
        refEndDate = request.GET.get('refEndDate')
        identifier = request.GET.get("identifier")
        if identifier:
            queryset = queryset.filter(chf_id=identifier)
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')
            if refDate != None:
                isValidDate = True
                try :
                    datevar = datetime.datetime.strptime(refDate, "%Y-%m-%d").date()
                except ValueError :
                    isValidDate = False
                queryset = queryset.filter(validity_from__gte=datevar)
            if refEndDate != None:
                isValidDate = True
                try :
                    datevar = datetime.datetime.strptime(refEndDate, "%Y-%m-%d").date()
                except ValueError :
                    isValidDate = False
                queryset = queryset.filter(validity_from__lt=datevar)

        serializer = ContractSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def get_queryset(self):
        return Policy.get_queryset(None, self.request.user)


class MedicationViewSet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin, GenericViewSet):
    lookup_field = 'uuid'
    serializer_class = MedicationSerializer
    permission_classes = (FHIRApiMedicationPermissions,)

    def get_queryset(self):
        return Item.get_queryset(None, self.request.user)


class ConditionViewSet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin, mixins.CreateModelMixin, GenericViewSet):
    lookup_field = 'id'
    serializer_class = ConditionSerializer
    permission_classes = (FHIRApiConditionPermissions,)

    def get_queryset(self):
        return Diagnosis.get_queryset(None, self.request.user)


class ActivityDefinitionViewSet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin, GenericViewSet):
    lookup_field = 'uuid'
    serializer_class = ActivityDefinitionSerializer
    permission_classes = (FHIRApiActivityDefinitionPermissions,)

    def get_queryset(self):
        return Service.get_queryset(None, self.request.user)


class HealthcareServiceViewSet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin, GenericViewSet):
    lookup_field = 'uuid'
    serializer_class = HealthcareServiceSerializer
    permission_classes = (FHIRApiHealthServicePermissions,)

    def get_queryset(self):
        return HealthFacility.get_queryset(None, self.request.user)



import core
from dateutil import parser


class TimeUtils(object):

    @classmethod
    def now(cls):
        return core.datetime.datetime.now()

    @classmethod
    def date(cls):
        return core.datetime.datetime.date(cls.now())

    @classmethod
    def str_to_date(cls, str_value):
        if type(str_value) is not str:
            str_value = str(str_value)
        py_date = parser.parse(str_value+" 00:00:00")
        return core.datetime.datetime.from_ad_datetime(py_date)



import copy

from insuree.models import Family
from api_fhir_r4.converters import GroupConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers import BaseFHIRSerializer


class GroupSerializer(BaseFHIRSerializer):
    fhirConverter = GroupConverter()

    def create(self, validated_data):
        chf_id = validated_data.get('head_insuree_id')
        if Family.objects.filter(head_insuree_id=chf_id).count() > 0:
            raise FHIRException('Exists family with the provided head')
        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        return Family.objects.create(**copied_data)

    def update(self, instance, validated_data):
        return instance



from api_fhir_r4 import views
from django.urls import include, path
from rest_framework.routers import DefaultRouter
from openIMIS.openimisapps import *


imis_modules = openimis_apps()


router = DefaultRouter()
router.register(r'login', views.LoginView, basename="login")
# register endpojnt related to Location module if used
if 'location' in imis_modules:
    router.register(r'Location', views.LocationViewSet, basename="Location_R4")
    router.register(r'HealthcareService', views.HealthcareServiceViewSet, basename="HealthcareService_R4")
# register endpoint for insuree if used
if 'insuree' in imis_modules:
    router.register(r'Patient', views.InsureeViewSet, basename="Patient_R4")
    router.register(r'Group', views.GroupViewSet, basename="Group_R4")
    router.register(
        r'CoverageEligibilityRequest', views.CoverageEligibilityRequestViewSet, basename="CoverageEligibilityRequest_R4"
    )
# register endpoints related to medical module
if 'medical' in imis_modules:
    router.register(r'Medication', views.MedicationViewSet, basename="Medication_R4")
    router.register(r'Condition', views.ConditionViewSet, basename="Condition_R4")
    router.register(r'ActivityDefinition', views.ActivityDefinitionViewSet, basename="ActivityDefinition_R4")
# register all endpoints related to claim based on claim
if 'claim' in imis_modules:
    router.register(r'Claim', views.ClaimViewSet, basename="Claim_R4")
    router.register(r'ClaimResponse', views.ClaimResponseViewSet, basename="ClaimResponse_R4")
    router.register(r'PractitionerRole', views.PractitionerRoleViewSet, basename="PractitionerRole_R4")
    router.register(r'Practitioner', views.PractitionerViewSet, basename="Practitioner_R4")
    router.register(r'CommunicationRequest', views.CommunicationRequestViewSet, basename="CommunicationRequest_R4")
# register endpoint for policy if used
if 'policy' in imis_modules:
    router.register(r'Coverage', views.CoverageRequestQuerySet, basename="Coverage_R4")
    router.register(r'Contract', views.ContractViewSet, basename="Contract_R4")
# register endpoint for policy holder if used
if 'policyholder' in imis_modules:
    router.register(r'Organisation', views.OrganisationViewSet, basename="Organisation_R4")
urlpatterns = [path('', include(router.urls))]



from policy.services import EligibilityRequest
from api_fhir_r4.configurations import R4CoverageEligibilityConfiguration as Config
from api_fhir_r4.converters import BaseFHIRConverter, PatientConverter
from fhir.resources.money import Money
from fhir.resources.coverageeligibilityresponse import CoverageEligibilityResponse as FHIRCoverageEligibilityResponse, \
    CoverageEligibilityResponseInsuranceItem, CoverageEligibilityResponseInsurance, CoverageEligibilityResponseInsuranceItemBenefit
from api_fhir_r4.models import CoverageEligibilityRequestV2 as FHIRCoverageEligibilityRequest
from api_fhir_r4.utils import TimeUtils


class CoverageEligibilityRequestConverter(BaseFHIRConverter):

    @classmethod
    def to_fhir_obj(cls, coverage_eligibility_response):
        fhir_response = FHIRCoverageEligibilityResponse.construct()
        cls.build_fhir_insurance(fhir_response, coverage_eligibility_response)
        return fhir_response

    @classmethod
    def to_imis_obj(cls, fhir_coverage_eligibility_request, audit_user_id):
        fhir_coverage_eligibility_request["status"] = "active"
        fhir_coverage_eligibility_request["purpose"] = ["validation"]
        fhir_coverage_eligibility_request["created"] = TimeUtils.date().isoformat()
        fhir_coverage_eligibility_request = FHIRCoverageEligibilityRequest(**fhir_coverage_eligibility_request)
        chf_id = cls.build_imis_uuid(fhir_coverage_eligibility_request)
        service_code = cls.build_imis_service_code(fhir_coverage_eligibility_request)
        item_code = cls.build_imis_item_code(fhir_coverage_eligibility_request)
        return EligibilityRequest(chf_id, service_code, item_code)

    @classmethod
    def build_fhir_insurance(cls, fhir_response, response):
        result = CoverageEligibilityResponseInsurance.construct()
        cls.build_fhir_int_item(result, Config.get_fhir_total_admissions_code(), response.total_admissions_left)
        cls.build_fhir_int_item(result, Config.get_fhir_total_visits_code(), response.total_visits_left)
        cls.build_fhir_int_item(result, Config.get_fhir_total_consultations_code(),
                                   response.total_consultations_left)
        cls.build_fhir_int_item(result, Config.get_fhir_total_surgeries_code(), response.total_surgeries_left)
        cls.build_fhir_int_item(result, Config.get_fhir_total_deliveries_code(), response.total_deliveries_left)
        cls.build_fhir_int_item(result, Config.get_fhir_total_antenatal_code(), response.total_antenatal_left)
        cls.build_fhir_money_item(result, Config.get_fhir_consultation_amount_code(),
                                     response.consultation_amount_left)
        cls.build_fhir_money_item(result, Config.get_fhir_surgery_amount_code(), response.surgery_amount_left)
        cls.build_fhir_money_item(result, Config.get_fhir_delivery_amount_code(), response.delivery_amount_left)
        cls.build_fhir_money_item(result, Config.get_fhir_hospitalization_amount_code(),
                                     response.hospitalization_amount_left)
        cls.build_fhir_money_item(result, Config.get_fhir_antenatal_amount_code(), response.antenatal_amount_left)
        cls.build_fhir_int_item(result, Config.get_fhir_service_left_code(), response.service_left)
        cls.build_fhir_int_item(result, Config.get_fhir_item_left_code(), response.item_left)
        cls.build_fhir_bool_item(result, Config.get_fhir_is_service_ok_code(), response.is_service_ok)
        cls.build_fhir_bool_item(result, Config.get_fhir_is_item_ok_code(), response.is_item_ok)
        if type(fhir_response.insurance) is not list:
            fhir_response.insurance = [result]
        else:
            fhir_response.insurance.append(result)

    @classmethod
    def build_fhir_bool_item(cls, insurance, code, is_ok):
        if is_ok is not None:
            item = cls.build_fhir_generic_item(code)
            item.excluded = not is_ok
            if type(insurance.item) is not list:
                insurance.item = [item]
            else:
                insurance.item.append(item)

    @classmethod
    def build_fhir_int_item(cls, insurance, code, value):
        if value is not None:
            item = cls.build_fhir_generic_item(code)
            cls.build_fhir_int_item_benefit(item, value)
            if type(insurance.item) is not list:
                insurance.item = [item]
            else:
                insurance.item.append(item)

    @classmethod
    def build_fhir_money_item(cls, insurance, code, value):
        if value is not None:
            item = cls.build_fhir_generic_item(code)
            cls.build_fhir_money_item_benefit(item, value)
            if type(insurance.item) is not list:
                insurance.item = [item]
            else:
                insurance.item.append(item)

    @classmethod
    def build_fhir_generic_item(cls, code):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        item.category = cls.build_simple_codeable_concept(code)
        return item

    @classmethod
    def build_fhir_int_item_benefit(cls, item, value):
        benefit = cls.build_fhir_generic_item_benefit()
        benefit.allowedUnsignedInt = value
        if type(item.benefit) is not list:
            item.benefit = [benefit]
        else:
            item.benefit.append(benefit)

    @classmethod
    def build_fhir_money_item_benefit(cls, item, value):
        benefit = cls.build_fhir_generic_item_benefit()
        money_value = Money.construct()
        money_value.value = value
        benefit.allowedMoney = money_value
        if type(item.benefit) is not list:
            item.benefit = [benefit]
        else:
            item.benefit.append(benefit)

    @classmethod
    def build_fhir_generic_item_benefit(cls):
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_simple_codeable_concept(Config.get_fhir_financial_code())
        return benefit

    @classmethod
    def build_imis_uuid(cls, fhir_coverage_eligibility_request):
        uuid = None
        patient_reference = fhir_coverage_eligibility_request.patient
        if patient_reference:
            uuid = PatientConverter.get_resource_id_from_reference(patient_reference)
        return uuid

    @classmethod
    def build_imis_service_code(cls, fhir_coverage_eligibility_request):
        return cls.get_text_from_codeable_concept_by_coding_code(fhir_coverage_eligibility_request.item[0].category,
                                                                 Config.get_fhir_service_code())

    @classmethod
    def build_imis_item_code(cls, fhir_coverage_eligibility_request):
        return cls.get_text_from_codeable_concept_by_coding_code(fhir_coverage_eligibility_request.item[0].productOrService,
                                                                 Config.get_fhir_item_code())

    @classmethod
    def get_text_from_codeable_concept_by_coding_code(cls, codeable_concept, coding_code):
        service_code = None
        if codeable_concept:
            coding = cls.get_first_coding_from_codeable_concept(codeable_concept)
            if coding and coding.code == coding_code:
                service_code = codeable_concept.text
        return service_code



from django.utils.translation import gettext
from location.models import Location

from api_fhir_r4.configurations import R4IdentifierConfig, R4LocationConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from fhir.resources.location import Location as FHIRLocation
from api_fhir_r4.models.imisModelEnums import ImisLocationType
from api_fhir_r4.utils import DbManagerUtils


class LocationConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_location, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_location = FHIRLocation.construct()
        cls.build_fhir_physical_type(fhir_location, imis_location)
        cls.build_fhir_pk(fhir_location, imis_location, reference_type)
        cls.build_fhir_location_identifier(fhir_location, imis_location)
        cls.build_fhir_location_name(fhir_location, imis_location)
        cls.build_fhir_type(fhir_location, imis_location)
        cls.build_fhir_part_of(fhir_location, imis_location, reference_type)
        cls.mode = 'instance'
        return fhir_location

    @classmethod
    def to_imis_obj(cls, fhir_location, audit_user_id):
        errors = []
        fhir_location = FHIRLocation(**fhir_location)
        imis_location = Location()
        cls.build_imis_location_identiftier(imis_location, fhir_location, errors)
        cls.build_imis_location_name(imis_location, fhir_location, errors)
        cls.build_imis_location_type(imis_location, fhir_location, errors)
        cls.build_imis_parent_location_id(imis_location, fhir_location, errors)
        cls.check_errors(errors)
        return imis_location

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_location_code_type()

    @classmethod
    def get_reference_obj_uuid(cls, imis_location: Location):
        return imis_location.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_location: Location):
        return imis_location.id

    @classmethod
    def get_reference_obj_code(cls, imis_location: Location):
        return imis_location.code


    @classmethod
    def get_fhir_resource_type(cls):
        return Location

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        location_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Location, uuid=location_uuid)

    @classmethod
    def build_fhir_location_identifier(cls, fhir_location, imis_location):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_location)
        fhir_location.identifier = identifiers

    @classmethod
    def build_fhir_location_code_identifier(cls, identifiers, imis_location):
        if imis_location is not None:
            identifier = cls.build_fhir_identifier(imis_location.code,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_location_code_type())
            identifiers.append(identifier)

    @classmethod
    def build_imis_location_identiftier(cls, imis_location, fhir_location, errors):
        value = cls.get_fhir_identifier_by_code(fhir_location.identifier,
                                                R4IdentifierConfig.get_fhir_location_code_type())
        if value:
            imis_location.code = value
        cls.valid_condition(imis_location.code is None, gettext('Missing location code'), errors)

    @classmethod
    def build_fhir_location_name(cls, fhir_location, imis_location):
        fhir_location.name = imis_location.name

    @classmethod
    def build_imis_location_name(cls, imis_location, fhir_location, errors):
        name = fhir_location.name
        if not cls.valid_condition(name is None,
                                   gettext('Missing location `name` attribute'), errors):
            imis_location.name = name

    @classmethod
    def build_fhir_physical_type(cls, fhir_location, imis_location):
        code = R4LocationConfig.get_fhir_code_for_area()
        text = "Area"
        fhir_location.physicalType = \
            cls.build_codeable_concept(code, R4LocationConfig.get_fhir_location_physical_type_system(), text=text)

    @classmethod
    def build_fhir_type(cls, fhir_location, imis_location):
        code = ""
        text = ""
        if imis_location.type == ImisLocationType.REGION.value:
            code = R4LocationConfig.get_fhir_code_for_region()
            text = "region"
        elif imis_location.type == ImisLocationType.DISTRICT.value:
            code = R4LocationConfig.get_fhir_code_for_district()
            text = "district"
        elif imis_location.type == ImisLocationType.WARD.value:
            code = R4LocationConfig.get_fhir_code_for_ward()
            text = "ward"
        elif imis_location.type == ImisLocationType.VILLAGE.value:
            code = R4LocationConfig.get_fhir_code_for_village()
            text = "village"
            
        fhir_location.type = \
            [cls.build_codeable_concept(code, R4LocationConfig.get_fhir_location_area_type_system(), text=text)]
        

    @classmethod
    def build_imis_location_type(cls, imis_location, fhir_location, errors):
        # get the type of location code
        code = fhir_location.type[0].coding[0].code
        if code == R4LocationConfig.get_fhir_code_for_region():
            imis_location.type = ImisLocationType.REGION.value
        elif code == R4LocationConfig.get_fhir_code_for_district():
            imis_location.type = ImisLocationType.DISTRICT.value
        elif code == R4LocationConfig.get_fhir_code_for_ward():
            imis_location.type = ImisLocationType.WARD.value
        elif code == R4LocationConfig.get_fhir_code_for_village():
            imis_location.type = ImisLocationType.VILLAGE.value
        cls.valid_condition(imis_location.type is None, gettext('Missing location type'), errors)

    @classmethod
    def build_fhir_part_of(cls, fhir_location, imis_location, reference_type):
        partOf = None
        if imis_location.parent is not None:
            fhir_location.partOf = LocationConverter.build_fhir_resource_reference(
                imis_location.parent, 'Location', imis_location.parent.code, reference_type=reference_type)

    @classmethod
    def build_imis_parent_location_id(cls, imis_location, fhir_location, errors):
        if fhir_location.partOf:
            parent_id = fhir_location.partOf
            if not cls.valid_condition(parent_id is None,
                                       gettext('Missing location `parent id` attribute'), errors):

                # get the imis parent location object, check if exists
                uuid_location = parent_id.identifier.value
                parent_location = Location.objects.filter(uuid=uuid_location)
                if parent_location:
                    parent_location = parent_location.first()
                    imis_location.parent = parent_location



from django.utils.translation import gettext
from location.models import HealthFacility, Location, HealthFacilityCatchment

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig, R4LocationConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.converters.locationConverter import LocationConverter
from fhir.resources.healthcareservice import HealthcareService as FHIRHealthcareService
from api_fhir_r4.models.imisModelEnums import ImisHfLevel
from api_fhir_r4.utils import TimeUtils, DbManagerUtils


class HealthcareServiceConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_hf, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_hcs = FHIRHealthcareService.construct()
        cls.build_fhir_pk(fhir_hcs, imis_hf, reference_type)
        cls.build_fhir_healthcare_service_identifier(fhir_hcs, imis_hf)
        cls.build_fhir_healthcare_service_name(fhir_hcs, imis_hf)
        cls.build_fhir_healthcare_service_category(fhir_hcs, imis_hf)
        cls.build_fhir_healthcare_service_extra_details(fhir_hcs, imis_hf)
        cls.build_fhir_healthcare_service_telecom(fhir_hcs, imis_hf)
        cls.build_fhir_location_reference(fhir_hcs, imis_hf, reference_type=reference_type)
        cls.build_fhir_healthcare_service_program(fhir_hcs, imis_hf)
        cls.build_fhir_healthcare_service_specialty(fhir_hcs, imis_hf)
        cls.build_fhir_healthcare_service_type(fhir_hcs, imis_hf)
        cls.build_fhir_healthcare_service_coverage_area(fhir_hcs, imis_hf, reference_type=reference_type)
        return fhir_hcs

    @classmethod
    def to_imis_obj(cls, fhir_hcs, audit_user_id):
        errors = []
        fhir_hcs = FHIRHealthcareService(**fhir_hcs)
        imis_hf = cls.createDefaultHealthFacility(audit_user_id)
        cls.build_imis_hf_identiftier(imis_hf, fhir_hcs, errors)
        cls.build_imis_location_identiftier(imis_hf, fhir_hcs, errors)
        cls.build_imis_hf_name(imis_hf, fhir_hcs, errors)
        cls.build_imis_hf_level(imis_hf, fhir_hcs, errors)
        cls.build_imis_hf_address(imis_hf, fhir_hcs)
        cls.build_imis_hf_contacts(imis_hf, fhir_hcs)
        cls.build_imis_legal_form(imis_hf, fhir_hcs, errors)
        cls.build_imis_sub_level(imis_hf, fhir_hcs, errors)
        cls.build_imis_care_type(imis_hf, fhir_hcs, errors)
        cls.check_errors(errors)
        return imis_hf

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_facility_id_type()

    @classmethod
    def get_reference_obj_uuid(cls, imis_hf: HealthFacility):
        return imis_hf.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_hf: HealthFacility):
        return imis_hf.id

    @classmethod
    def get_reference_obj_code(cls, imis_hf: HealthFacility):
        return imis_hf.code

    @classmethod
    def get_fhir_resource_type(cls):
        return FHIRHealthcareService

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        healthfacility_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(HealthFacility, uuid=healthfacility_uuid)

    @classmethod
    def createDefaultHealthFacility(cls, audit_user_id):
        imis_hf = HealthFacility()
        # TODO legalForm isn't covered because that value is missing in the model (value need to be nullable in DB)
        # TODO LocationId isn't covered because that value is missing in the model (value need to be nullable in DB)
        imis_hf.offline = GeneralConfiguration.get_default_value_of_location_offline_attribute()
        imis_hf.care_type = GeneralConfiguration.get_default_value_of_location_care_type()
        imis_hf.validity_from = TimeUtils.now()
        imis_hf.audit_user_id = audit_user_id
        return imis_hf

    @classmethod
    def build_fhir_healthcare_service_identifier(cls, fhir_location, imis_hf):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_hf)
        fhir_location.identifier = identifiers

    @classmethod
    def build_fhir_hf_code_identifier(cls, identifiers, imis_hf):
        if imis_hf is not None:
            identifier = cls.build_fhir_identifier(imis_hf.code,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_facility_id_type())
            identifiers.append(identifier)

    @classmethod
    def build_imis_hf_identiftier(cls, imis_hf, fhir_location, errors):
        value = cls.get_fhir_identifier_by_code(fhir_location.identifier,
                                                R4IdentifierConfig.get_fhir_facility_id_type())
        if value:
            imis_hf.code = value
        cls.valid_condition(imis_hf.code is None, gettext('Missing hf code'), errors)

    @classmethod
    def build_imis_location_identiftier(cls, imis_hf, fhir_location, errors):
        value = cls.get_fhir_identifier_by_code(fhir_location.identifier,
                                                R4IdentifierConfig.get_fhir_facility_id_type())
        # TODO - fix assigning location in HF
        if value:
            imis_hf.location.code = value
        cls.valid_condition(imis_hf.location.code is None, gettext('Missing location code'), errors)

    @classmethod
    def build_fhir_healthcare_service_name(cls, fhir_hcs, imis_hf):
        fhir_hcs.name = imis_hf.name

    @classmethod
    def build_imis_hf_name(cls, imis_hf, fhir_hcs, errors):
        name = fhir_hcs.name
        if not cls.valid_condition(name is None,
                                   gettext('Missing patient `name` attribute'), errors):
            imis_hf.name = name

    @classmethod
    def build_fhir_healthcare_service_category(cls, fhir_hcs, imis_hf):
        code = ""
        text = ""
        if imis_hf.level == ImisHfLevel.HEALTH_CENTER.value:
            code = R4LocationConfig.get_fhir_code_for_health_center()
            text = "Hospital center"
        elif imis_hf.level == ImisHfLevel.HOSPITAL.value:
            code = R4LocationConfig.get_fhir_code_for_hospital()
            text = "Hospital"
        elif imis_hf.level == ImisHfLevel.DISPENSARY.value:
            code = R4LocationConfig.get_fhir_code_for_dispensary()
            text = "Dispensary"

        fhir_hcs.category = \
            [cls.build_codeable_concept(code, R4LocationConfig.get_fhir_location_site_type_system(), text=text)]

    @classmethod
    def build_imis_hf_level(cls, imis_hf, fhir_hcs, errors):
        location_type = fhir_hcs.category
        if not cls.valid_condition(location_type is None,
                                   gettext('Missing patient `type` attribute'), errors):
            for maritalCoding in location_type.coding:
                if maritalCoding.system == R4LocationConfig.get_fhir_location_site_type_system():
                    code = maritalCoding.code
                    if code == R4LocationConfig.get_fhir_code_for_health_center():
                        imis_hf.level = ImisHfLevel.HEALTH_CENTER.value
                    elif code == R4LocationConfig.get_fhir_code_for_hospital():
                        imis_hf.level = ImisHfLevel.HOSPITAL.value
                    elif code == R4LocationConfig.get_fhir_code_for_dispensary():
                        imis_hf.level = ImisHfLevel.DISPENSARY.value

            cls.valid_condition(imis_hf.level is None, gettext('Missing hf level'), errors)

    @classmethod
    def build_fhir_healthcare_service_extra_details(cls, fhir_hcs, imis_hf):
        fhir_hcs.extraDetails = imis_hf.address

    @classmethod
    def build_imis_hf_address(cls, imis_hf, fhir_hcs):
        address = fhir_hcs.extraDetails
        if address is not None:
            if address.type == 'physical':
                imis_hf.address = address.text

    @classmethod
    def build_fhir_healthcare_service_telecom(cls, fhir_hcs, imis_hf):
        telecom = []
        if imis_hf.phone is not None and imis_hf.phone != "":
            phone = HealthcareServiceConverter.build_fhir_contact_point(imis_hf.phone, 'phone',
                                                               "home")
            telecom.append(phone)
        if imis_hf.fax is not None and imis_hf.fax != "":
            fax = HealthcareServiceConverter.build_fhir_contact_point(imis_hf.fax, 'fax',
                                                             "home")
            telecom.append(fax)
        if imis_hf.email is not None and imis_hf.email != "":
            email = HealthcareServiceConverter.build_fhir_contact_point(imis_hf.email, 'email',
                                                               "home")
            telecom.append(email)
        fhir_hcs.telecom = telecom

    @classmethod
    def build_imis_hf_contacts(cls, imis_hf, fhir_hcs):
        telecom = fhir_hcs.telecom
        if telecom is not None:
            for contact_point in telecom:
                if contact_point.system == 'phone':
                    imis_hf.phone = contact_point.value
                elif contact_point.system == 'fax':
                    imis_hf.fax = contact_point.value
                elif contact_point.system == 'email':
                    imis_hf.email = contact_point.value

    @classmethod
    def build_fhir_location_reference(cls, fhir_hcs, imis_hf, reference_type):
        if imis_hf.location is not None:
            fhir_hcs.location = [
                LocationConverter.build_fhir_resource_reference(imis_hf.location, reference_type=reference_type)]

    @classmethod
    def build_fhir_healthcare_service_program(cls, fhir_hcs, imis_hf):
        fhir_hcs.program = [cls.build_codeable_concept(imis_hf.legal_form.code, text=imis_hf.legal_form.legal_form)]

    @classmethod
    def build_imis_legal_form(cls, imis_hf, fhir_hcs, errors):
        code = fhir_hcs.program.coding
        text = fhir_hcs.program.text
        if not cls.valid_condition(code and text is None,
                                   gettext('Missing healthcare service `legal form` attribute'), errors):
            imis_hf.legal_form.code = code
            imis_hf.legal_form.legal_form = text

    @classmethod
    def build_fhir_healthcare_service_specialty(cls, fhir_hcs, imis_hf):
        if imis_hf.sub_level is not None:
            fhir_hcs.specialty = [cls.build_codeable_concept(imis_hf.sub_level.code,
                                                             text=imis_hf.sub_level.health_facility_sub_level)]

    @classmethod
    def build_imis_sub_level(cls, imis_hf, fhir_hcs, errors):
        code = fhir_hcs.specialty.coding
        text = fhir_hcs.specialty.text
        if not cls.valid_condition(code and text is None,
                                   gettext('Missing healthcare service `legal form` attribute'), errors):
            imis_hf.sub_level.code = code
            imis_hf.sub_level.health_facility_sub_level = text

    @classmethod
    def build_fhir_healthcare_service_type(cls, fhir_hcs, imis_hf):
        code = ""
        text = ""
        if imis_hf.care_type == HealthFacility.CARE_TYPE_IN_PATIENT:
            code = R4LocationConfig.get_fhir_code_for_in_patient()
            text = "In-patient"
        elif imis_hf.care_type == HealthFacility.CARE_TYPE_OUT_PATIENT:
            code = R4LocationConfig.get_fhir_code_for_out_patient()
            text = "Out-patient"
        elif imis_hf.care_type == HealthFacility.CARE_TYPE_BOTH:
            code = R4LocationConfig.get_fhir_code_for_both()
            text = "Both"

        fhir_hcs.type = \
            [cls.build_codeable_concept(code, R4LocationConfig.get_fhir_hf_service_type_system(), text=text)]

    @classmethod
    def build_imis_care_type(cls, imis_hf, fhir_hcs, errors):
        care_type = fhir_hcs.type
        if not cls.valid_condition(care_type is None,
                                   gettext('Missing hf `care type` attribute'), errors):
            for maritalCoding in care_type.coding:
                if maritalCoding.system == R4LocationConfig.get_fhir_hf_service_type_system():
                    code = maritalCoding.code
                    if code == R4LocationConfig.get_fhir_code_for_in_patient():
                        imis_hf.care_type = HealthFacility.CARE_TYPE_IN_PATIENT
                    elif code == R4LocationConfig.get_fhir_code_for_out_patient():
                        imis_hf.care_type = HealthFacility.CARE_TYPE_OUT_PATIENT
                    elif code == R4LocationConfig.get_fhir_code_for_both():
                        imis_hf.care_type = HealthFacility.CARE_TYPE_BOTH

            cls.valid_condition(imis_hf.care_type is None, gettext('Missing hf care type'), errors)

    @classmethod
    def build_fhir_healthcare_service_coverage_area(cls, fhir_hcs, imis_hf, reference_type):
        imis_hf_catchments = HealthFacilityCatchment.objects\
            .filter(health_facility=imis_hf) \
            .select_related("location")\
            .values("location")
        for catchment in imis_hf_catchments:
            location = Location.objects.filter(id=catchment["location"])
            fhir_hcs.coverageArea.append(
                LocationConverter.build_fhir_resource_reference(location[0], reference_type=reference_type))



from api_fhir_r4.configurations import BaseConfiguration
from api_fhir_r4.defaultConfig import DEFAULT_CFG


class GeneralConfiguration(BaseConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        config = cls.get_config()
        config.default_audit_user_id = cfg['default_audit_user_id']
        config.gender_codes = cfg['gender_codes']
        config.default_value_of_patient_head_attribute = cfg['default_value_of_patient_head_attribute']
        config.default_value_of_patient_card_issued_attribute = cfg['default_value_of_patient_card_issued_attribute']
        config.default_value_of_location_offline_attribute = cfg['default_value_of_location_offline_attribute']
        config.default_value_of_location_care_type = cfg['default_value_of_location_care_type']
        config.default_response_page_size = cfg['default_response_page_size']

    @classmethod
    def get_default_audit_user_id(cls):
        return cls.get_config_attribute("default_audit_user_id")

    @classmethod
    def get_male_gender_code(cls):
        return cls.get_config_attribute("gender_codes").get('male', 'M')

    @classmethod
    def get_female_gender_code(cls):
        return cls.get_config_attribute("gender_codes").get('female', 'F')

    @classmethod
    def get_other_gender_code(cls):
        return cls.get_config_attribute("gender_codes").get('other', 'O')

    @classmethod
    def get_default_value_of_patient_head_attribute(cls):
        return cls.get_config_attribute("default_value_of_patient_head_attribute")

    @classmethod
    def get_default_value_of_patient_card_issued_attribute(cls):
        return cls.get_config_attribute("default_value_of_patient_card_issued_attribute")

    @classmethod
    def get_default_value_of_location_offline_attribute(cls):
        return cls.get_config_attribute("default_value_of_location_offline_attribute")

    @classmethod
    def get_default_value_of_location_care_type(cls):
        return cls.get_config_attribute("default_value_of_location_care_type")

    @classmethod
    def get_default_response_page_size(cls):
        return cls.get_config_attribute("default_response_page_size")

    @classmethod
    def show_system(cls):
        return 0



from medical.models import Service
from fhir.resources.activitydefinition import ActivityDefinition
from fhir.resources.extension import Extension
from fhir.resources.money import Money
from fhir.resources.usagecontext import UsageContext
from fhir.resources.codeableconcept import CodeableConcept
from fhir.resources.coding import Coding
from api_fhir_r4.converters import R4IdentifierConfig, BaseFHIRConverter, ReferenceConverterMixin
from django.utils.translation import gettext

from api_fhir_r4.models.imisModelEnums import ImisCategoryDefinition
from api_fhir_r4.utils import DbManagerUtils, TimeUtils
import core


class ActivityDefinitionConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_activity_definition, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_activity_definition = ActivityDefinition.construct()
        # first to construct is status - obligatory fields
        cls.build_fhir_status(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_pk(fhir_activity_definition, imis_activity_definition, reference_type)
        cls.build_fhir_identifiers(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_date(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_name(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_title(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_use_context(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_topic(fhir_activity_definition, imis_activity_definition)
        cls.build_activity_definition_extension(fhir_activity_definition, imis_activity_definition)
        cls.build_fhir_frequency_extension(fhir_activity_definition, imis_activity_definition)
        return fhir_activity_definition

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_service_code_type()

    @classmethod
    def to_imis_obj(cls, fhir_activity_definition, audit_user_id):
        errors = []
        fhir_activity_definition = ActivityDefinition(**fhir_activity_definition)
        imis_activity_definition = Service()
        cls.build_imis_identifier(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_validity_from(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_code(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_name(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_type(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_pat_cat(imis_activity_definition, fhir_activity_definition)
        cls.build_imis_serv_category(imis_activity_definition, fhir_activity_definition, errors)
        cls.build_imis_serv_care_type(imis_activity_definition, fhir_activity_definition, errors)
        cls.check_errors(errors)
        return imis_activity_definition

    @classmethod
    def get_reference_obj_uuid(cls, imis_activity_definition: Service):
        return imis_activity_definition.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_activity_definition: Service):
        return imis_activity_definition.id

    @classmethod
    def get_reference_obj_code(cls, imis_activity_definition: Service):
        return imis_activity_definition.code

    @classmethod
    def get_fhir_resource_type(cls):
        return ActivityDefinition

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_activity_definition_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Service, code=imis_activity_definition_code)

    @classmethod
    def build_fhir_identifiers(cls, fhir_activity_definition, imis_activity_definition):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_activity_definition)
        fhir_activity_definition.identifier = identifiers

    @classmethod
    def build_imis_identifier(cls, imis_activity_definition, fhir_activity_definition, errors):
        value = cls.get_fhir_identifier_by_code(fhir_activity_definition.identifier,
                                                R4IdentifierConfig.get_fhir_uuid_type_code())
        if value:
            imis_activity_definition.code = value
        cls.valid_condition(imis_activity_definition.code is None, gettext('Missing the service code'), errors)

    @classmethod
    def build_fhir_status(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.status = "active"

    @classmethod
    def build_fhir_date(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.date = imis_activity_definition.validity_from.isoformat()

    @classmethod
    def build_imis_validity_from(cls, imis_activity_definition, fhir_activity_definition, errors):
        validity_from = fhir_activity_definition.date
        if not cls.valid_condition(validity_from is None,
                                   gettext('Missing activity definition `validity from` attribute'), errors):
            imis_activity_definition.validity_from = TimeUtils.str_to_date(validity_from)

    @classmethod
    def build_fhir_name(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.name = imis_activity_definition.code

    @classmethod
    def build_imis_serv_code(cls, imis_activity_definition, fhir_activity_definition, errors):
        serv_code = fhir_activity_definition.name
        if not cls.valid_condition(serv_code is None,
                                   gettext('Missing activity definition `serv code` attribute'), errors):
            imis_activity_definition.code = serv_code

    @classmethod
    def build_fhir_title(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.title = imis_activity_definition.name

    @classmethod
    def build_imis_serv_name(cls, imis_activity_definition, fhir_activity_definition, errors):
        serv_name = fhir_activity_definition.title
        if not cls.valid_condition(serv_name is None,
                                   gettext('Missing activity definition `serv name` attribute'), errors):
            imis_activity_definition.name = serv_name
   
    @classmethod
    def build_imis_serv_pat_cat(cls, imis_activity_definition, fhir_activity_definition):
        serv_pat_cat = fhir_activity_definition.useContext.code
        number = 0
        if "K" in serv_pat_cat:
            number = number + 8
        if "A" in serv_pat_cat:
            number = number + 4
        if "F" in serv_pat_cat:
            number = number + 2
        if "M" in serv_pat_cat:
            number = number + 1

        imis_activity_definition.patient_category = number

    @classmethod
    def build_imis_serv_category(cls, imis_activity_definition, fhir_activity_definition, errors):
        serv_category = fhir_activity_definition.useContext.text
        if not cls.valid_condition(serv_category is None,
                                   gettext('Missing activity definition `serv category` attribute'), errors):
            imis_activity_definition.category = serv_category

    @classmethod
    def build_imis_serv_care_type(cls, imis_activity_definition, fhir_activity_definition, errors):
        serv_care_type = fhir_activity_definition.useContext.text
        if not cls.valid_condition(serv_care_type is None,
                                   gettext('Missing activity definition `serv care type` attribute'), errors):
            imis_activity_definition.care_type = serv_care_type

    @classmethod
    def build_fhir_topic(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.topic = [cls.build_codeable_concept(
            "DefinitionTopic", "http://terminology.hl7.org/CodeSystem/definition-topic",
            text=imis_activity_definition.type)]

    @classmethod
    def build_imis_serv_type(cls, imis_activity_definition, fhir_activity_definition, errors):
        serv_type = fhir_activity_definition.topic
        if not cls.valid_condition(serv_type is None,
                                   gettext('Missing activity definition `serv type` attribute'), errors):
            imis_activity_definition.topic = serv_type

    @classmethod
    def build_fhir_code(cls, fhir_activity_definition, imis_activity_definition):
        fhir_activity_definition.code = cls.build_codeable_concept(imis_activity_definition.code,
                                                                   text=imis_activity_definition.name)

    @classmethod
    def build_activity_definition_extension(cls, fhir_activity_definition, imis_activity_definition):
        cls.build_unit_price(fhir_activity_definition, imis_activity_definition)
        return fhir_activity_definition

    @classmethod
    def build_unit_price(cls, fhir_activity_definition, imis_activity_definition):
        unit_price = cls.build_unit_price_extension(imis_activity_definition.price)
        if type(fhir_activity_definition.extension) is not list:
            fhir_activity_definition.extension = [unit_price]
        else:
            fhir_activity_definition.extension.append(unit_price)

    @classmethod
    def build_unit_price_extension(cls, value):
        extension = Extension.construct()
        money = Money.construct()
        extension.url = "unitPrice"
        extension.valueMoney = money
        extension.valueMoney.value = value
        if hasattr(core, 'currency'):
            extension.valueMoney.currency = core.currency
        return extension

    @classmethod
    def build_fhir_frequency_extension(cls, fhir_activity_definition, imis_activity_definition):
        serv_price = cls.build_fhir_serv_frequency_extension(imis_activity_definition)
        if type(fhir_activity_definition.extension) is not list:
            fhir_activity_definition.extension = [serv_price]
        else:
            fhir_activity_definition.extension.append(serv_price)

    @classmethod
    def build_fhir_serv_frequency_extension(cls, imis_activity_definition):
        extension = Extension.construct()
        extension.url = "frequency"
        extension.valueInteger = imis_activity_definition.frequency
        return extension

    @classmethod
    def build_fhir_use_context(cls, fhir_activity_definition, imis_activity_definition):
        use_context = cls.build_fhir_use_context_context(imis_activity_definition)
        fhir_activity_definition.useContext = use_context

    @classmethod
    def build_fhir_use_context_context(cls, imis_activity_definition):
        usage_context_gender = \
            cls.__build_usage_context('useContextGender', cls.build_fhir_gender(imis_activity_definition))
        usage_context_age = \
            cls.__build_usage_context('useContextAge', cls.build_fhir_age(imis_activity_definition))
        usage_context_workflow = \
            cls.__build_usage_context('useContextWorkflow',  cls.build_fhir_workflow(imis_activity_definition))
        usage_context_venue = \
            cls.__build_usage_context('useContextVenue', cls.build_fhir_venue(imis_activity_definition))
        usage_context_level = \
            cls.__build_usage_context('useContextLevel', cls.build_fhir_level(imis_activity_definition))

        usage = [usage_context_gender, usage_context_age]
        if usage_context_workflow.valueCodeableConcept.coding[0].display:
            usage.append(usage_context_workflow)
        elif usage_context_venue.valueCodeableConcept.coding[0].display:
            usage.append(usage_context_venue)

        usage.append(usage_context_level)

        for x in usage:
            a = all([a.code for a in x.valueCodeableConcept.coding])
            if a:
                print(x)
        return usage

    @classmethod
    def __build_usage_context(cls, code, codeable_concept):
        usage_context = UsageContext.construct()
        usage_context.valueCodeableConcept = CodeableConcept.construct()
        usage_context.code = Coding.construct()
        usage_context.code.code = code
        usage_context.valueCodeableConcept = codeable_concept
        return usage_context

    @classmethod
    def build_fhir_gender(cls, imis_activity_definition):
        male = cls.build_fhir_male(imis_activity_definition)
        female = cls.build_fhir_female(imis_activity_definition)

        codeable_concept = CodeableConcept.construct()

        if male:
            coding_male = Coding.construct()
            coding_male.code = male
            coding_male.display = "Male"
            if type(codeable_concept.coding) is not list:
                codeable_concept.coding = [coding_male]
            else:
                codeable_concept.coding.append(coding_male)

        if female:
            coding_female = Coding.construct()
            coding_female.code = female
            coding_female.display = "Female"
            if type(codeable_concept.coding) is not list:
                codeable_concept.coding = [coding_female]
            else:
                codeable_concept.coding.append(coding_female)
            codeable_concept.text = "Male or Female"

        return codeable_concept

    @classmethod
    def build_fhir_age(cls, imis_activity_definition):
        adult = cls.build_fhir_adult(imis_activity_definition)
        kid = cls.build_fhir_kid(imis_activity_definition)
        codeable_concept = CodeableConcept.construct()

        if adult:
            coding_adult = Coding.construct()
            coding_adult.code = adult
            coding_adult.display = "Adult"
            if type(codeable_concept.coding) is not list:
                codeable_concept.coding = [coding_adult]
            else:
                codeable_concept.coding.append(coding_adult)
            codeable_concept.text = "Adult"

        if kid:
            coding_kid = Coding.construct()
            coding_kid.code = kid
            coding_kid.display = "Kid"
            if type(codeable_concept.coding) is not list:
                codeable_concept.coding = [coding_kid]
            else:
                codeable_concept.coding.append(coding_kid)
            codeable_concept.text = "Adult or Kid"

        return codeable_concept

    @classmethod
    def build_fhir_venue(cls, imis_activity_definition):
        display = ""
        if imis_activity_definition.care_type == "O":
            display = "Out-patient"
        if imis_activity_definition.care_type == "I":
            display = "In-patient"
        if imis_activity_definition.care_type == "B":
            display = "Both"

        codeable_concept = CodeableConcept.construct()
        coding_venue = Coding.construct()
        coding_venue.code = imis_activity_definition.care_type
        coding_venue.display = display
        if type(codeable_concept.coding) is not list:
            codeable_concept.coding = [coding_venue]
        else:
            codeable_concept.coding.append(coding_venue)
        codeable_concept.text = "Clinical Venue"
        return codeable_concept

    @classmethod
    def build_fhir_level(self, imis_activity_definition: Service):
        # Values for this extension are fixed for medication
        display = ""
        if imis_activity_definition.level == 'S':
            display = 'Simple Service'
        elif imis_activity_definition.level == 'V':
            display = 'Visit'
        elif imis_activity_definition.level == 'D':
            display = 'Day of stay'
        elif imis_activity_definition.level == 'H':
            display = 'Hospital case'

        codeable_concept = CodeableConcept.construct()
        coding_level = Coding.construct()
        coding_level.code = imis_activity_definition.level
        coding_level.display = display
        if type(codeable_concept.coding) is not list:
            codeable_concept.coding = [coding_level]
        else:
            codeable_concept.coding.append(coding_level)
        codeable_concept.text = "Service Level"

        return codeable_concept

    @classmethod
    def build_fhir_workflow(cls, imis_activity_definition):
        codeable_concept = CodeableConcept.construct()
        coding_workflow = Coding.construct()

        if imis_activity_definition.category and imis_activity_definition.category != ' ':
            coding_workflow.code = imis_activity_definition.category
            coding_workflow.display = ImisCategoryDefinition.get_category_display(imis_activity_definition.category)

        if type(codeable_concept.coding) is not list:
            codeable_concept.coding = [coding_workflow]
        else:
            codeable_concept.coding.append(coding_workflow)
        codeable_concept.text = "Workflow Setting"
        return codeable_concept

    @classmethod
    def build_fhir_male(cls, imis_activity_definition):
        item_pat_cat = imis_activity_definition.patient_category
        male = ""
        if item_pat_cat > 8:
            kid = "K"
            item_pat_cat = item_pat_cat - 8
        if item_pat_cat > 4:
            adult = "A"
            item_pat_cat = item_pat_cat - 4
        if item_pat_cat > 2:
            female = "F"
            item_pat_cat = item_pat_cat - 2
        if item_pat_cat == 1:
            male = "M"
        return male

    @classmethod
    def build_fhir_female(cls, imis_activity_definition):
        item_pat_cat = imis_activity_definition.patient_category
        female = ""
        if item_pat_cat > 8:
            kid = "K"
            item_pat_cat = item_pat_cat - 8
        if item_pat_cat > 4:
            adult = "A"
            item_pat_cat = item_pat_cat - 4
        if item_pat_cat >= 2:
            female = "F"
        return female

    @classmethod
    def build_fhir_adult(cls, imis_activity_definition):
        item_pat_cat = imis_activity_definition.patient_category
        adult = ""
        if item_pat_cat > 8:
            kid = "K"
            item_pat_cat = item_pat_cat - 8
        if item_pat_cat >= 4:
            adult = "A"
        return adult

    @classmethod
    def build_fhir_kid(cls, imis_activity_definition):
        item_pat_cat = imis_activity_definition.patient_category
        kid = ""
        if item_pat_cat >= 8:
            kid = "K"
        return kid



from insuree.models import Gender
from rest_framework.test import APITestCase

from api_fhir_r4.models import Patient
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin, FhirApiCreateTestMixin, \
    FhirApiUpdateTestMixin, FhirApiDeleteTestMixin


class PatientAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, FhirApiCreateTestMixin, FhirApiUpdateTestMixin,
                      FhirApiDeleteTestMixin, APITestCase):

    base_url = '/api_fhir_r4/Patient/'
    _test_json_path = "/test/test_patient.json"
    _TEST_GENDER_CODE = 'M'
    _TEST_EXPECTED_NAME = "UPDATED_NAME"

    def setUp(self):
        super(PatientAPITests, self).setUp()

    def verify_updated_obj(self, updated_obj):
        self.assertTrue(isinstance(updated_obj, Patient))
        self.assertEqual(self._TEST_EXPECTED_NAME, updated_obj.name[0].family)

    def update_resource(self, data):
        data['name'][0]['family'] = self._TEST_EXPECTED_NAME

    def create_dependencies(self):
        gender = Gender()
        gender.code = self._TEST_GENDER_CODE
        gender.save()



from claim.models import ClaimAdmin

from api_fhir_r4.configurations import R4IdentifierConfig
from api_fhir_r4.converters import PractitionerConverter
from api_fhir_r4.models import HumanName, NameUse, Identifier, ContactPoint, ContactPointSystem, Practitioner, \
    ContactPointUse
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import TimeUtils


class PractitionerTestMixin(GenericTestMixin):

    _TEST_LAST_NAME = "Smith"
    _TEST_OTHER_NAME = "John"
    _TEST_DOB = "1990-03-24T00:00:00"
    _TEST_ID = 1
    _TEST_UUID = "254f6268-964b-4d8d-aa26-20081f22235e"
    _TEST_CODE = "1234abcd"
    _TEST_PHONE = "813-996-476"
    _TEST_EMAIL = "TEST@TEST.com"

    def create_test_imis_instance(self):
        imis_claim_admin = ClaimAdmin()
        imis_claim_admin.last_name = self._TEST_LAST_NAME
        imis_claim_admin.other_names = self._TEST_OTHER_NAME
        imis_claim_admin.id = self._TEST_ID
        imis_claim_admin.uuid = self._TEST_UUID
        imis_claim_admin.code = self._TEST_CODE
        imis_claim_admin.dob = TimeUtils.str_to_date(self._TEST_DOB)
        imis_claim_admin.phone = self._TEST_PHONE
        imis_claim_admin.email_id = self._TEST_EMAIL
        return imis_claim_admin

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_LAST_NAME, imis_obj.last_name)
        self.assertEqual(self._TEST_OTHER_NAME, imis_obj.other_names)
        self.assertEqual(self._TEST_CODE, imis_obj.code)
        expected_date = TimeUtils.str_to_date(self._TEST_DOB)
        self.assertEqual(expected_date, imis_obj.dob)
        self.assertEqual(self._TEST_PHONE, imis_obj.phone)
        self.assertEqual(self._TEST_EMAIL, imis_obj.email_id)

    def create_test_fhir_instance(self):
        fhir_practitioner = Practitioner()
        name = HumanName()
        name.family = self._TEST_LAST_NAME
        name.given = [self._TEST_OTHER_NAME]
        name.use = NameUse.USUAL.value
        fhir_practitioner.name = [name]
        identifiers = []
        chf_id = PractitionerConverter.build_fhir_identifier(self._TEST_CODE,
                                                             R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                             R4IdentifierConfig.get_fhir_claim_admin_code_type())
        identifiers.append(chf_id)
        fhir_practitioner.identifier = identifiers
        fhir_practitioner.birthDate = self._TEST_DOB
        telecom = []
        phone = PractitionerConverter.build_fhir_contact_point(self._TEST_PHONE, ContactPointSystem.PHONE.value,
                                                               ContactPointUse.HOME.value)
        telecom.append(phone)
        email = PractitionerConverter.build_fhir_contact_point(self._TEST_EMAIL, ContactPointSystem.EMAIL.value,
                                                               ContactPointUse.HOME.value)
        telecom.append(email)
        fhir_practitioner.telecom = telecom
        return fhir_practitioner

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(1, len(fhir_obj.name))
        human_name = fhir_obj.name[0]
        self.assertTrue(isinstance(human_name, HumanName))
        self.assertEqual(self._TEST_OTHER_NAME, human_name.given[0])
        self.assertEqual(self._TEST_LAST_NAME, human_name.family)
        self.assertEqual(NameUse.USUAL.value, human_name.use)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = PractitionerConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_claim_admin_code_type():
                self.assertEqual(self._TEST_CODE, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(self._TEST_UUID, identifier.value)
        self.assertEqual(self._TEST_DOB, fhir_obj.birthDate)
        self.assertEqual(2, len(fhir_obj.telecom))
        for telecom in fhir_obj.telecom:
            self.assertTrue(isinstance(telecom, ContactPoint))
            if telecom.system == ContactPointSystem.PHONE.value:
                self.assertEqual(self._TEST_PHONE, telecom.value)
            elif telecom.system == ContactPointSystem.EMAIL.value:
                self.assertEqual(self._TEST_EMAIL, telecom.value)



import datetime
import os
from functools import lru_cache
from unittest import mock, TestCase

from claim.models import ClaimItem, ClaimService
from location.models import HealthFacility
from medical.models import Diagnosis, Item, Service
from prompt_toolkit.cache import memoized

from api_fhir_r4.converters import ClaimResponseConverter, ConditionConverter, HealthcareServiceConverter
from api_fhir_r4.models import FHIRBaseObject, ClaimResponseError, CodeableConcept, Coding, Extension, Period, \
    ClaimResponsePayment
from api_fhir_r4.tests import ClaimResponseTestMixin


class ClaimResponseConverterTestCase(ClaimResponseTestMixin):

    __TEST_CLAIM_RESPONSE_JSON_PATH = "/test/test_claimResponse.json"

    def setUp(self):
        super(ClaimResponseConverterTestCase, self).setUp()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self._test_claim_response_json_representation = open(dir_path + self.__TEST_CLAIM_RESPONSE_JSON_PATH).read()
        if self._test_claim_response_json_representation[-1:] == "\n":
            self._test_claim_response_json_representation = self._test_claim_response_json_representation[:-1]

    @mock.patch('claim.models.ClaimItem.objects')
    @mock.patch('claim.models.ClaimService.objects')
    def test_to_fhir_obj(self, cs_mock, ci_mock):
        cs_mock.filter.return_value = [self._TEST_SERVICE]
        ci_mock.filter.return_value = [self._TEST_ITEM]

        imis_claim_response = self.create_test_imis_instance()
        item = self.create_test_claim_item()
        item.claim = imis_claim_response
        item.item.save()
        item.save()
        service = self.create_test_claim_service()
        service.claim = imis_claim_response
        service.service.save()
        service.save()
        fhir_claim_response = ClaimResponseConverter.to_fhir_obj(imis_claim_response)
        self.verify_fhir_instance(fhir_claim_response)

    # def test_fhir_object_to_json_request(self):
    #     fhir_obj = self.create_test_fhir_instance()
    #     actual_representation = fhir_obj.dumps(format_='json')
    #     self.assertEqual(self._test_claim_response_json_representation, actual_representation)

    # def test_create_object_from_json(self):
    #     fhir_claim = FHIRBaseObject.loads(self._test_claim_response_json_representation, 'json')
    #     self.verify_fhir_instance(fhir_claim)

    def test_fhir_object_to_imis_object(self):
        expected_claim, fhir_obj = self.get_or_create_test_data()
        actual_claim = ClaimResponseConverter.to_imis_obj(fhir_obj, 1)
        self.assertEqual(expected_claim, actual_claim)

    def test_build_imis_outcome(self):
        claim, fhir_obj = self.get_or_create_test_data()
        fhir_obj.outcome = 'checked'
        ClaimResponseConverter.build_imis_outcome(claim, fhir_obj)
        self.assertEqual(claim.status, 4)

    def test_build_imis_errors(self):
        claim, fhir_obj = self.get_or_create_test_data()
        test_error = ClaimResponseError()
        test_error.code = self.build_test_codeable_concept(str(10))
        fhir_obj.error = [test_error]
        ClaimResponseConverter.build_imis_errors(claim, fhir_obj)
        self.assertEqual(claim.rejection_reason, 10)

    def test_build_imis_type(self):
        claim, fhir_obj = self.get_or_create_test_data()
        fhir_obj.type = self.build_test_codeable_concept(None, 'E')
        ClaimResponseConverter.build_imis_type(claim, fhir_obj)
        self.assertEqual(claim.visit_type, 'E')

    def test_build_imis_status(self):
        claim, fhir_obj = self.get_or_create_test_data()
        fhir_obj.status = "Not Selected"
        ClaimResponseConverter.build_imis_status(claim, fhir_obj)
        self.assertEqual(claim.review_status, 2)

    def test_build_imis_requestor(self):
        claim, fhir_obj = self.get_or_create_test_data()
        hf = HealthFacility.objects.get(id=2)
        fhir_obj.requestor = HealthcareServiceConverter.build_fhir_resource_reference(hf)
        ClaimResponseConverter.build_imis_requestor(claim, fhir_obj)
        self.assertEqual(claim.health_facility, hf)

    def test_build_imis_billable_period(self):
        claim, fhir_obj = self.get_or_create_test_data()
        date_from = datetime.date(2020, 10, 5)
        date_to = datetime.date(2020, 11, 5)
        fhir_obj.extension = [self.build_test_billable_period_extension_concept(date_from, date_to)]
        ClaimResponseConverter.build_imis_billable_period(claim, fhir_obj)
        self.assertEqual(claim.date_from, date_from)
        self.assertEqual(claim.date_to, date_to)

    def test_build_imis_diagnoses(self):
        claim, fhir_obj = self.get_or_create_test_data()
        d1 = self.build_test_diagonosis('D1')
        d1.save()
        d2 = self.build_test_diagonosis('D2')
        d2.save()
        fhir_obj.extension = [
            self.build_diagnosis_extensions('icd_0', d1),
            self.build_diagnosis_extensions('icd_1', d2),
        ]
        ClaimResponseConverter.build_imis_diagnoses(claim, fhir_obj)
        self.assertEqual(claim.icd, d1)
        self.assertEqual(claim.icd_1, d2)

    def test_build_imis_adjustment(self):
        claim, fhir_obj = self.get_or_create_test_data()
        adjustment = 'TestAdjustment1'
        fhir_obj.payment = self.build_test_adjustment(adjustment)
        ClaimResponseConverter.build_imis_adjustment(claim, fhir_obj)
        self.assertEqual(claim.adjustment, adjustment)

    def test_build_imis_items(self):
        claim, fhir_obj = self.get_or_create_test_data()

        claim_items = list(ClaimItem.objects.filter(claim=claim))
        claim_services = list(ClaimService.objects.filter(claim=claim))

        ClaimResponseConverter.build_imis_items(claim, fhir_obj)

        self.assertListEqual(claim.claim_items, claim_items)
        self.assertListEqual(claim.claim_services, claim_services)

    @lru_cache(maxsize=None)
    def get_or_create_test_data(self):
        imis_instance = self.create_test_imis_instance()
        imis_instance.health_facility = HealthFacility.objects.get(id=1)
        imis_instance.icd.save()
        imis_instance.save()

        item = self.build_claim_item(imis_instance)
        item.save()

        service = self.build_claim_service(imis_instance)
        service.save()

        fhir_instance = ClaimResponseConverter.to_fhir_obj(imis_instance)
        return imis_instance, fhir_instance

    def build_test_codeable_concept(self, code=None, text=None):
        codeable_concept = CodeableConcept()
        if code:
            coding = Coding()
            if not isinstance(code, str):
                code = str(code)
            coding.code = code
            codeable_concept.coding = [coding]
        if text:
            codeable_concept.text = text
        return codeable_concept

    def build_test_billable_period_extension_concept(self, date_from=None, date_to=None):
        extension = Extension()
        extension.url = "billablePeriod"
        extension.valuePeriod = Period()
        if date_from:
            extension.valuePeriod.start = date_from.isoformat()
        if date_to:
            extension.valuePeriod.end = date_to.isoformat()
        return extension

    def build_diagnosis_extensions(self, icd_order, diagnosis):
        extension = Extension()
        extension.url = icd_order
        extension.valueReference = ConditionConverter.build_fhir_resource_reference(diagnosis)
        return extension

    def build_test_adjustment(self, adjustment):
        fhir_payment = ClaimResponsePayment()
        fhir_payment.adjustmentReason = self.build_test_codeable_concept(None, adjustment)
        return fhir_payment

    def build_test_diagonosis(self, code):
        d = Diagnosis(code=code)
        d.audit_user_id = self._ADMIN_AUDIT_USER_ID
        return d

    def build_claim_item(self, claim):
        item = Item.objects.get(id=1)
        claim_item = ClaimItem()
        claim_item.claim = claim
        claim_item.item = item
        claim_item.status = 4
        claim_item.availability = True
        claim_item.audit_user_id = 1
        self.assign_default_qty(claim_item)
        self.assign_default_prices(claim_item)
        return claim_item

    def build_claim_service(self, claim):
        service = Service.objects.get(id=1)
        claim_service = ClaimService()
        claim_service.claim = claim
        claim_service.service = service
        claim_service.status = 2
        claim_service.audit_user_id = 1
        self.assign_default_qty(claim_service)
        self.assign_default_prices(claim_service)
        return claim_service

    def assign_default_qty(self, claim_item):
        claim_item.qty_provided = 10
        claim_item.qty_approved = 9

    def assign_default_prices(self, claim_item):
        claim_item.price_asked = 100
        claim_item.price_approved = 90
        claim_item.price_adjusted = 90
        claim_item.price_valuated = 90



import os
from unittest import mock

from api_fhir_r4.converters import LocationConverter, PatientConverter, PractitionerConverter
from api_fhir_r4.converters.claimConverter import ClaimConverter
from api_fhir_r4.models import FHIRBaseObject
from api_fhir_r4.tests import ClaimTestMixin


class ClaimConverterTestCase(ClaimTestMixin):

    __TEST_CLAIM_JSON_PATH = "/test/test_claim.json"

    def setUp(self):
        super(ClaimConverterTestCase, self).setUp()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self._test_claim_json_representation = open(dir_path + self.__TEST_CLAIM_JSON_PATH).read()
        if self._test_claim_json_representation[-1:] == "\n":
            self._test_claim_json_representation = self._test_claim_json_representation[:-1]

    @mock.patch('claim.models.ClaimItem.objects')
    @mock.patch('claim.models.ClaimService.objects')
    def test_to_fhir_obj(self, cs_mock, ci_mock):
        cs_mock.filter.return_value = [self._TEST_SERVICE]
        ci_mock.filter.return_value = [self._TEST_ITEM]

        imis_claim = self.create_test_imis_instance()
        fhir_claim = ClaimConverter.to_fhir_obj(imis_claim)
        self.verify_fhir_instance(fhir_claim)

    @mock.patch.object(LocationConverter, 'get_imis_obj_by_fhir_reference')
    @mock.patch.object(PractitionerConverter, 'get_imis_obj_by_fhir_reference')
    @mock.patch('medical.models.Diagnosis.objects')
    @mock.patch.object(PatientConverter, 'get_imis_obj_by_fhir_reference')
    def test_to_imis_obj(self, mock_insuree, mock_cdc, mock_pc, mock_hfc):
        self.setUp()
        mock_insuree.return_value = self._TEST_INSUREE
        mock_cdc.get.return_value = self._TEST_DIAGNOSIS_CODE
        mock_pc.return_value = self._TEST_CLAIM_ADMIN
        mock_hfc.return_value = self._TEST_HF

        fhir_claim = self.create_test_fhir_instance()
        imis_claim = ClaimConverter.to_imis_obj(fhir_claim, None)
        self.verify_imis_instance(imis_claim)

    def test_fhir_object_to_json_request(self):
        self.setUp()
        fhir_obj = self.create_test_fhir_instance()
        actual_representation = fhir_obj.dumps(format_='json')
        self.assertEqual(self._test_claim_json_representation, actual_representation)

    def test_create_object_from_json(self):
        self.setUp()
        fhir_claim = FHIRBaseObject.loads(self._test_claim_json_representation, 'json')
        self.verify_fhir_instance(fhir_claim)



from claim.models import Feedback

from api_fhir_r4.configurations import R4CommunicationRequestConfig as Config
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from fhir.resources.communicationrequest import CommunicationRequest
from api_fhir_r4.utils import DbManagerUtils


class CommunicationRequestConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_feedback, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_communication_request = CommunicationRequest.construct()
        fhir_communication_request.status = "unknown"
        cls.build_fhir_occurrence_datetime(fhir_communication_request, imis_feedback)
        cls.build_fhir_pk(fhir_communication_request, imis_feedback, reference_type)
        cls.build_fhir_identifiers(fhir_communication_request, imis_feedback)
        cls.build_fhir_reason_codes(fhir_communication_request, imis_feedback)
        cls.build_fhir_status(fhir_communication_request)
        return fhir_communication_request

    @classmethod
    def get_reference_obj_id(cls, imis_feedback):
        return imis_feedback.uuid

    @classmethod
    def get_fhir_resource_type(cls):
        return CommunicationRequest

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_feedback_id = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Feedback, pk=imis_feedback_id)

    @classmethod
    def build_fhir_occurrence_datetime(cls, fhir_communication_request, imis_feedback):
        feedback_date = imis_feedback.feedback_date
        if feedback_date:
            fhir_communication_request.occurrenceDateTime = feedback_date.isoformat()

    @classmethod
    def build_fhir_identifiers(cls, fhir_communication_request, imis_feedback):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_feedback)
        fhir_communication_request.identifier = identifiers

    @classmethod
    def build_all_identifiers(cls, identifiers, imis_object):
        # Feedback does not provide code reference
        cls.build_fhir_uuid_identifier(identifiers, imis_object)
        cls.build_fhir_id_identifier(identifiers, imis_object)

    @classmethod
    def build_fhir_reason_codes(cls, fhir_communication_request, imis_feedback):
        reasons = [cls.build_codeable_concept(Config.get_fhir_care_rendered_code(),
                                              text=str(imis_feedback.care_rendered)),
                   cls.build_codeable_concept(Config.get_fhir_payment_asked_code(),
                                              text=str(imis_feedback.payment_asked)),
                   cls.build_codeable_concept(Config.get_fhir_drug_prescribed_code(),
                                              text=str(imis_feedback.drug_prescribed)),
                   cls.build_codeable_concept(Config.get_fhir_drug_received_code(),
                                              text=str(imis_feedback.drug_received))]
        if imis_feedback.asessment is not None:
            reasons.append(cls.build_codeable_concept(Config.get_fhir_asessment_code(),
                                                      text=str(imis_feedback.asessment)))
        fhir_communication_request.reasonCode = reasons

    @classmethod
    def build_fhir_status(cls, fhir_claim_response):
        fhir_claim_response.status = "active"



from ctypes import Union
from typing import List, Tuple
from django.db import models
from collections.abc import Iterable

from api_fhir_r4.converters import ReferenceConverterMixin
from api_fhir_r4.exceptions import FHIRException
from fhir.resources.fhirabstractmodel import FHIRAbstractModel


class ContainedResourceConverter:
    """It's used for extracting IMIS Model attributes in FHIR format.

    Methods
    ----------
    convert_from_source(imis_obj: models.Model):
        Convert IMIS Model attribute to FHIR Object
    """

    def __init__(self, imis_resource_name, resource_fhir_converter, resource_extract_method=None,
                 reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        """
        Parameters
        ----------
        :param imis_resource_name: Name of attribute, which value should be transformed.
        :param resource_fhir_converter: FHIR Converter, it's used for mapping imis resource to fhir representation.
        It must implement at least to_fhir_obj() method.
        :param resource_extract_method: Optional argument. Function used for getting attribute value from IMIS Model.
        It has two arguments, first is django model, second one is imis_resource_name. Default function is
        imis_model.__getattribute__(imis_resource_name). Return type can be model or iterable (e.g. list of attributes).
        :param reference_type: Optional argument. Determine what object value will be used as reference and id.
        """
        self.imis_resource_name = imis_resource_name
        self.extract_value = resource_extract_method or (lambda model, attribute: model.__getattribute__(attribute))
        self.converter = resource_fhir_converter()
        self.reference_type = reference_type

    def convert_from_source(self, imis_obj: models.Model) -> List[FHIRAbstractModel]:
        """Convert IMIS Model attribute to FHIR Object.

        :param imis_obj: IMIS Object with attribute that have to be converted.
        :return: Attribute converted to FHIR object list. If attribute is single object then it's still converted to
        list format.
        """
        resource = self.extract_value(imis_obj, self.imis_resource_name)
        if isinstance(resource, Iterable):
            return self._convert_all_values(list(resource))
        else:
            converted = self._convert_single_resource(resource)
            return [converted] if converted else []

    def _convert_all_values(self, imis_resources: List[object]) -> List[FHIRAbstractModel]:
        return [self._convert_single_resource(next_resource) for next_resource in imis_resources]

    def _convert_single_resource(self, imis_resource_value: object) -> FHIRAbstractModel:
        if imis_resource_value is not None:
            try:
                fhir_value = self.converter.to_fhir_obj(imis_resource_value, self.reference_type)
                return fhir_value
            except Exception as e:
                raise FHIRException("Failed to process: {}, exception: \n{}"
                                    .format((self.imis_resource_name, imis_resource_value), str(e)))



import json
import os

from core.models import User
from core.services import create_or_update_interactive_user, create_or_update_core_user

from rest_framework import status
from rest_framework.test import APITestCase
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiCreateTestMixin
from api_fhir_r4.utils import DbManagerUtils


class LoginAPITests(GenericFhirAPITestMixin, FhirApiCreateTestMixin, APITestCase):

    base_url = '/api_fhir_r4/login/'
    _test_json_path = "/test/test_login.json"
    _test_json_path_wrong_credentials = "/tests/test/test_login_bad_credentials.json"
    _test_json_path_wrong_payload = "/tests/test/test_login_bad_payload.json"
    _TEST_EXPECTED_NAME = "UPDATED_NAME"
    _TEST_USER_NAME = "TestUserTest2"
    _TEST_USER_PASSWORD = "TestPasswordTest2"
    _TEST_DATA_USER = {
        "username": _TEST_USER_NAME,
        "last_name": _TEST_USER_NAME,
        "password": _TEST_USER_PASSWORD,
        "other_names": _TEST_USER_NAME,
        "user_types": "INTERACTIVE",
        "language": "en",
        "roles": [9],
    }
    _test_request_data_wrong_credentials = None
    _test_request_data_bad_payload = None

    def setUp(self):
        super(LoginAPITests, self).setUp()
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        json_representation = open(dir_path + self._test_json_path_wrong_credentials).read()
        self._test_request_data_wrong_credentials = json.loads(json_representation)
        json_representation = open(dir_path + self._test_json_path_wrong_payload).read()
        self._test_request_data_bad_payload = json.loads(json_representation)

    def get_or_create_user_api(self):
        user = DbManagerUtils.get_object_or_none(User, username=self._TEST_USER_NAME)
        if user is None:
            user = self.__create_user_interactive_core()
        return user

    def get_bundle_from_json_response(self, response):
        pass

    def __create_user_interactive_core(self):
        i_user, i_user_created = create_or_update_interactive_user(
            user_id=None, data=self._TEST_DATA_USER, audit_user_id=999, connected=False)
        create_or_update_core_user(
            user_uuid=None, username=self._TEST_DATA_USER["username"], i_user=i_user)
        return DbManagerUtils.get_object_or_none(User, username=self._TEST_USER_NAME)

    def test_post_should_create_correctly(self):
        self.get_or_create_user_api()
        response = self.client.post(self.base_url, data=self._test_request_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_post_should_give_unathorized(self):
        response = self.client.post(self.base_url, data=self._test_request_data_wrong_credentials, format='json')
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_post_should_give_bad_request(self):
        response = self.client.post(self.base_url, data=self._test_request_data_bad_payload, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_get_should_required_login(self):
        pass
